# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Widget that handles communications between a console in debugging
mode and Spyder
"""

# Standard library imports
import atexit
import pdb
import re

# Third-party imports
from IPython.core.history import HistoryManager
from IPython.core.inputtransformer2 import TransformerManager
from IPython.lib.lexers import (
    IPython3Lexer, Python3Lexer, bygroups, using
)
from pygments.token import Keyword, Operator
from pygments.util import ClassNotFound
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtpy.QtCore import QEvent
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.config.base import get_conf_path


class SpyderIPy3Lexer(IPython3Lexer):
    # Detect !cmd command and highlight them
    tokens = IPython3Lexer.tokens
    spyder_tokens = [
        (r'(!)(\w+)(.*\n)', bygroups(Operator, Keyword, using(Python3Lexer))),
        (r'(%)(\w+)(.*\n)', bygroups(Operator, Keyword, using(Python3Lexer))),
    ]
    tokens['root'] = spyder_tokens + tokens['root']


class PdbHistory(HistoryManager):

    def _get_hist_file_name(self, profile=None):
        """
        Get default pdb history file name.

        The profile parameter is ignored, but must exist for compatibility with
        the parent class.
        """
        return get_conf_path('pdb_history.sqlite')


class DebuggingHistoryWidget(RichJupyterWidget):
    """
    Widget with the necessary attributes and methods to override the pdb
    history mechanism while debugging.
    """
    PDB_HIST_MAX = 400

    def __init__(self, *args, **kwargs):
        # History
        self._pdb_history_input_number = 0  # Input number for current session
        self._saved_pdb_history_input_number = {}  # for recursive debugging

        # Catch any exception that prevents to create or access the history
        # file to avoid errors.
        # Fixes spyder-ide/spyder#18531
        try:
            self._pdb_history_file = PdbHistory()
            self._pdb_history = [
                line[-1] for line in self._pdb_history_file.get_tail(
                    self.PDB_HIST_MAX, include_latest=True)
            ]
        except Exception:
            self._pdb_history_file = None
            self._pdb_history = []

        self._pdb_history_edits = {}  # Temporary history edits
        self._pdb_history_index = len(self._pdb_history)

        # super init
        super(DebuggingHistoryWidget, self).__init__(*args, **kwargs)

    # --- Public API --------------------------------------------------
    def new_history_session(self):
        """Start a new history session."""
        self._pdb_history_input_number = 0
        if self._pdb_history_file is not None:
            self._pdb_history_file.new_session()

    def end_history_session(self):
        """End an history session."""
        self._pdb_history_input_number = 0
        if self._pdb_history_file is not None:
            self._pdb_history_file.end_session()

    def add_to_pdb_history(self, line):
        """Add command to history"""
        self._pdb_history_input_number += 1
        line_num = self._pdb_history_input_number
        self._pdb_history_index = len(self._pdb_history)
        self._pdb_history_edits = {}
        line = line.rstrip()
        if not line:
            return

        # If repeated line
        history = self._pdb_history
        if len(history) > 0 and history[-1] == line:
            return

        cmd = line.split(" ")[0]
        args = line.split(" ")[1:]
        is_pdb_cmd = (
            cmd.strip() and cmd[0] != '!' and "do_" + cmd in dir(pdb.Pdb)
        )
        if self.is_pdb_using_exclamantion_mark():
            is_pdb_cmd = is_pdb_cmd or (
                cmd.strip() and cmd[0] == '!'
                and "do_" + cmd[1:] in dir(pdb.Pdb))

        if cmd and (not is_pdb_cmd or len(args) > 0):
            self._pdb_history.append(line)
            self._pdb_history_index = len(self._pdb_history)
            if self._pdb_history_file is not None:
                self._pdb_history_file.store_inputs(line_num, line)

    # --- Private API (overrode by us) --------------------------------
    @property
    def _history(self):
        """Get history."""
        if self.is_debugging():
            return self._pdb_history
        else:
            return self.__history

    @_history.setter
    def _history(self, history):
        """Set history."""
        if self.is_debugging():
            self._pdb_history = history
        else:
            self.__history = history

    @property
    def _history_edits(self):
        """Get edited history."""
        if self.is_debugging():
            return self._pdb_history_edits
        else:
            return self.__history_edits

    @_history_edits.setter
    def _history_edits(self, history_edits):
        """Set edited history."""
        if self.is_debugging():
            self._pdb_history_edits = history_edits
        else:
            self.__history_edits = history_edits

    @property
    def _history_index(self):
        """Get history index."""
        if self.is_debugging():
            return self._pdb_history_index
        else:
            return self.__history_index

    @_history_index.setter
    def _history_index(self, history_index):
        """Set history index."""
        if self.is_debugging():
            self._pdb_history_index = history_index
        else:
            self.__history_index = history_index


class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
    """
    Widget with the necessary attributes and methods to handle
    communications between a console in debugging mode and
    Spyder
    """

    CONF_SECTION = 'ipython_console'

    def __init__(self, *args, **kwargs):
        # Communication state
        self._pdb_recursion_level = 0  # Number of debbuging loop we are in
        self._pdb_input_ready = False  # Can we send a command now
        self._waiting_pdb_input = False  # Are we waiting on the user
        # Other state
        self._pdb_prompt = None  # prompt
        self._pdb_prompt_input = False  # wether pdb waits for input or comm
        self._pdb_last_cmd = ''  # last command sent to pdb
        self._pdb_frame_loc = (None, None)  # fname, lineno
        self._pdb_take_focus = True  # Focus to shell after command execution
        # Command queue
        self._pdb_input_queue = []  # List of (code, hidden, echo_stack_entry)
        # Temporary flags
        self._tmp_reading = False
        # super init
        super(DebuggingWidget, self).__init__(*args, **kwargs)

        # Adapted from qtconsole/frontend_widget.py
        # This adds the IPdb as a prompt self._highlighter recognises
        self._highlighter._ipy_prompt_re = re.compile(
            r'^({})?('.format(re.escape(self.other_output_prefix)) +
            r'[ \t]*\(*IPdb \[\d+\]\)*: |' +
            r'[ \t]*In \[\d+\]: |[ \t]*\ \ \ \.\.\.+: )')

        # Reset debug state when debugging is done
        self.sig_prompt_ready.connect(self.reset_debug_state)

    # --- Public API --------------------------------------------------

    def shutdown(self):
        """
        Close the save thread and database file.
        """
        try:
            # Make sure the database will not be called after closing
            self.sig_prompt_ready.disconnect(self.reset_debug_state)
        except (TypeError, RuntimeError):
            # Already disconnected
            pass

        if self._pdb_history_file is not None:
            try:
                self._pdb_history_file.save_thread.stop()
                # Now that it was called, no need to call it at exit
                atexit.unregister(self._pdb_history_file.save_thread.stop)
            except AttributeError:
                pass

            try:
                self._pdb_history_file.db.close()
            except AttributeError:
                pass

    # --- Comm API --------------------------------------------------

    def set_debug_state(self, recursion_level):
        """Update the debug state."""
        if recursion_level == self._pdb_recursion_level:
            # Nothing to change
            return

        if recursion_level > self._pdb_recursion_level:
            # Start debugging
            if self._pdb_recursion_level > 0:
                # Recursive debugging, save state
                self._saved_pdb_history_input_number[
                    self._pdb_recursion_level] = self._pdb_history_input_number
                self.end_history_session()
            self.new_history_session()
        elif recursion_level < self._pdb_recursion_level:
            # Stop debugging
            self.end_history_session()
            if recursion_level > 0:
                # Still debugging, restore state
                self.new_history_session()
                self._pdb_history_input_number = (
                    self._saved_pdb_history_input_number.pop(
                        recursion_level, 0))

        # If debugging starts or stops, clear the input queue.
        self._pdb_recursion_level = recursion_level
        self._pdb_input_queue = []
        self._pdb_frame_loc = (None, None)

    def _pdb_cmd_prefix(self):
        """Return the command prefix"""
        prefix = ''
        if self.spyder_kernel_ready and self.is_pdb_using_exclamantion_mark():
            prefix = '!'
        return prefix

    def pdb_execute_command(self, command):
        """
        Execute a pdb command
        """
        self._pdb_take_focus = False
        self.pdb_execute(
            self._pdb_cmd_prefix() + command, hidden=False,
            echo_stack_entry=False, add_history=False)

    def _handle_input_request(self, msg):
        """Process an input request."""
        if not self.is_spyder_kernel and "ipdb>" in msg['content']['prompt']:
            # Check if we can guess a path from the shell content:
            self._flush_pending_stream()
            cursor = self._get_end_cursor()
            cursor.setPosition(self._prompt_pos, QTextCursor.KeepAnchor)
            text = cursor.selection().toPlainText()
            match = re.search(r"> (.*\.py)\((\d+)\)", text)
            state = None

            if match:
                fname, lineno = match.groups()
                state = {
                    'step': {
                        'fname': fname,
                        'lineno': int(lineno)
                    }
                }

            prompt = msg['content']['prompt']
            password = msg['content']['password']
            self.pdb_input(prompt, password, state, from_input=True)
            return
        return super(DebuggingWidget, self)._handle_input_request(msg)

    def pdb_execute(self, line, hidden=False, echo_stack_entry=True,
                    add_history=True):
        """
        Send line to the pdb kernel if possible.

        Parameters
        ----------
        line: str
            the line to execute

        hidden: bool
            If the line should be hidden

        echo_stack_entry: bool
            If not hidden, if the stack entry should be printed

        add_history: bool
            If not hidden, wether the line should be added to history
        """
        if not self.is_debugging():
            return

        if not line.strip():
            # Must get the last genuine command
            line = self._pdb_last_cmd

        if hidden:
            # Don't show stack entry if hidden
            echo_stack_entry = False
        else:
            if not self.is_waiting_pdb_input():
                # We can't execute this if we are not waiting for pdb input
                self._pdb_input_queue.append(
                    (line, hidden, echo_stack_entry, add_history))
                return

            if line.strip():
                self._pdb_last_cmd = line

            # Print the text if it is programatically added.
            if line.strip() != self.input_buffer.strip():
                self.input_buffer = line
            self._append_plain_text('\n')

            if add_history:
                # Save history to browse it later
                self.add_to_pdb_history(line)

            # Set executing to true and save the input buffer
            self._input_buffer_executing = self.input_buffer
            self._executing = True
            self._waiting_pdb_input = False

            # Disable the console
            self._tmp_reading = False
            self._finalize_input_request()
            hidden = True

            # Emit executing
            self.executing.emit(line)
            self.sig_pdb_state_changed.emit(False)

        if self._pdb_input_ready:
            # Print the string to the console
            self._pdb_input_ready = False
            self.pdb_input_reply(line, echo_stack_entry)
            return

        self._pdb_input_queue.append(
            (line, hidden, echo_stack_entry, add_history))

    def reset_debug_state(self):
        """Reset debug state if the debugger crashed."""
        self.set_debug_state(0)

    # --- To Sort --------------------------------------------------
    def stop_debugging(self):
        """Stop debugging."""
        if self.spyder_kernel_ready and not self.is_waiting_pdb_input():
            self.interrupt_kernel()
        self.pdb_execute_command("exit")

    def is_pdb_using_exclamantion_mark(self):
        return self.get_conf('pdb_use_exclamation_mark', section='debugger')

    def refresh_from_pdb(self, pdb_state):
        """
        Refresh Variable Explorer and Editor from a Pdb session,
        after running any pdb command.

        See publish_pdb_state and notify_spyder in spyder_kernels
        """
        pdb_step = pdb_state.pop('step', None)
        if pdb_step and 'fname' in pdb_step:
            fname = pdb_step['fname']
            lineno = pdb_step['lineno']

            last_pdb_loc = self._pdb_frame_loc
            self._pdb_frame_loc = (fname, lineno)

            # Only step if the location changed
            if (fname, lineno) != last_pdb_loc:
                self.sig_pdb_step.emit(fname, lineno)

        if "do_where" in pdb_state:
            fname, lineno = self._pdb_frame_loc
            if fname:
                self.sig_pdb_step.emit(fname, lineno)

        pdb_stack = pdb_state.pop('stack', None)
        if pdb_stack:
            pdb_stack, pdb_index = pdb_stack
            self.sig_pdb_stack.emit(pdb_stack, pdb_index)

        request_pdb_input =  pdb_state.pop('request_pdb_input', None)
        if request_pdb_input:
            self.pdb_execute(request_pdb_input)

        self.update_state(pdb_state)

    def show_pdb_output(self, text):
        """Show Pdb output."""
        self._append_plain_text(self.output_sep, before_prompt=True)
        prompt = self._current_out_prompt()
        self._append_html(
            '<span class="out-prompt">%s</span>' % prompt,
            before_prompt=True
        )
        # If the repr is multiline, make sure we start on a new line,
        # so that its lines are aligned.
        if "\n" in text and not self.output_sep.endswith("\n"):
            self._append_plain_text('\n', before_prompt=True)
        self._append_plain_text(text + self.output_sep2, before_prompt=True)
        self._append_plain_text('\n', before_prompt=True)

    def get_pdb_last_step(self):
        """Get last pdb step retrieved from a Pdb session."""
        return self._pdb_frame_loc

    def is_debugging(self):
        """Check if we are debugging."""
        return self._pdb_recursion_level > 0

    def debugging_depth(self):
        """Debugging depth"""
        return self._pdb_recursion_level

    def is_waiting_pdb_input(self):
        """Check if we are waiting a pdb input."""
        # If the comm is not open, self._pdb_recursion_level can not be set
        return self.is_debugging() and self._waiting_pdb_input

    # ---- Public API (overrode by us) ----------------------------
    def reset(self, clear=False):
        """
        Resets the widget to its initial state if ``clear`` parameter
        is True
        """
        super(DebuggingWidget, self).reset(clear)
        # Make sure the prompt is printed
        if clear and self.is_waiting_pdb_input():
            prompt = self._pdb_prompt

            try:
                # This is necessary to avoid an error when the iopub channel is
                # closed.
                # See jupyter/qtconsole#574
                if not self.kernel_client.iopub_channel.closed():
                    self.kernel_client.iopub_channel.flush()
            except AttributeError:
                self.kernel_client.iopub_channel.flush()

            self._reading = False
            self._readline(prompt=prompt, callback=self.pdb_execute)

    # --- Private API --------------------------------------------------
    def _current_prompt(self):
        prompt = "IPdb [{}]".format(self._pdb_history_input_number + 1)
        for i in range(self._pdb_recursion_level - 1):
            # Add recursive debugger prompt
            prompt = "({})".format(prompt)
        return prompt + ": "

    def _current_out_prompt(self):
        """Get current out prompt."""
        prompt = "Out\u00A0\u00A0[{}]".format(self._pdb_history_input_number)
        for i in range(self._pdb_recursion_level - 1):
            # Add recursive debugger prompt
            prompt = "({})".format(prompt)
        return prompt + ": "

    def _handle_kernel_info_reply(self, rep):
        """Handle kernel info replies."""
        super(DebuggingWidget, self)._handle_kernel_info_reply(rep)
        pygments_lexer = rep['content']['language_info'].get(
            'pygments_lexer', '')
        try:
            # add custom lexer
            if pygments_lexer == 'ipython3':
                lexer = SpyderIPy3Lexer()
            else:
                return
            self._highlighter._lexer = lexer
        except ClassNotFound:
            pass

    def _redefine_complete_for_dbg(self, client):
        """Redefine kernel client's complete method to work while debugging."""

        original_complete = client.complete

        def complete(code, cursor_pos=None):
            if self.is_waiting_pdb_input():
                shell_channel = client.shell_channel
                client._shell_channel = client.control_channel
                try:
                    return original_complete(code, cursor_pos)
                finally:
                    client._shell_channel = shell_channel
            else:
                return original_complete(code, cursor_pos)

        client.complete = complete

    def _update_pdb_prompt(self, prompt):
        """Update the prompt that is recognised as a pdb prompt."""
        if prompt == self._pdb_prompt:
            # Nothing to do
            return

        self._pdb_prompt = prompt

        # Update continuation prompt to reflect (possibly) new prompt length.
        self._set_continuation_prompt(
            self._make_continuation_prompt(prompt), html=True)

    def _is_pdb_complete(self, source):
        """
        Check if the pdb input is ready to be executed.
        """
        if source and source[0] == '!':
            source = source[1:]
        tm = TransformerManager()
        complete, indent = tm.check_complete(source)
        if indent is not None:
            indent = indent * ' '
        return complete != 'incomplete', indent

    def execute(self, source=None, hidden=False, interactive=False):
        """
        Executes source or the input buffer, possibly prompting for more
        input.

        Do not use to run pdb commands (such as `continue`).
        Use pdb_execute instead. This will add a '!' in front of the code.
        """
        if self.is_waiting_pdb_input():
            if source is None:
                if hidden:
                    # Nothing to execute
                    return
                else:
                    source = self.input_buffer
            else:
                if not self.is_pdb_using_exclamantion_mark():
                    source = '!' + source
                if not hidden:
                    self.input_buffer = source

            if interactive:
                # Add a continuation prompt if not complete
                complete, indent = self._is_pdb_complete(source)
                if not complete:
                    self.do_execute(source, complete, indent)
                    return
            if hidden:
                self.pdb_execute(source, hidden)
            else:
                if self._reading_callback:
                    self._reading_callback()
            return

        return super(DebuggingWidget, self).execute(
            source, hidden, interactive)

    def pdb_input(self, prompt, password=None, state=None, from_input=False):
        """Get input for a command."""

        self.set_debug_state(1 + prompt.count("("))

        if state is not None and isinstance(state, dict):
            self.refresh_from_pdb(state)

        # Replace with numbered prompt
        prompt = self._current_prompt()
        self._update_pdb_prompt(prompt)
        self._pdb_prompt_input = from_input

        # The prompt should be printed unless:
        # 1. The prompt is already printed (self._reading is True)
        # 2. A hidden command is in the queue
        print_prompt = (not self._reading
                        and (len(self._pdb_input_queue) == 0
                             or not self._pdb_input_queue[0][1]))

        if print_prompt:
            # Make sure that all output from the SUB channel has been processed
            # before writing a new prompt.
            try:
                # This is necessary to avoid an error when the iopub channel is
                # closed.
                # See jupyter/qtconsole#574
                if not self.kernel_client.iopub_channel.closed():
                    self.kernel_client.iopub_channel.flush()
            except AttributeError:
                self.kernel_client.iopub_channel.flush()

            self._waiting_pdb_input = True
            self._readline(prompt=prompt,
                           callback=self.pdb_execute,
                           password=password)
            self._executing = False
            self._highlighter.highlighting_on = True
            # The previous code finished executing
            self.executed.emit(self._pdb_prompt)
            self.sig_pdb_prompt_ready.emit()
            self.sig_pdb_state_changed.emit(True)

        self._pdb_input_ready = True

        start_line = self.get_conf('startup/pdb_run_lines', default='')
        # Only run these lines when printing a new prompt
        if start_line and print_prompt and self.is_waiting_pdb_input():
            # Send a few commands
            self.pdb_execute(start_line, hidden=True)
            return

        # While the widget thinks only one input is going on,
        # other functions can be sending messages to the kernel.
        # This must be properly processed to avoid dropping messages.
        # If the kernel was not ready, the messages are queued.
        if len(self._pdb_input_queue) > 0:
            args = self._pdb_input_queue.pop(0)
            self.pdb_execute(*args)
            return

    def pdb_input_reply(self, line, echo_stack_entry):
        """Send a pdb input to the kernel."""
        if self._pdb_prompt_input:
            # Send line to input
            self.kernel_client.input(line)
            return

        self.call_kernel(interrupt=True).pdb_input_reply(
            line, echo_stack_entry=echo_stack_entry)

    # --- Private API (overrode by us) ----------------------------------------
    def _show_prompt(self, prompt=None, html=False, newline=True,
                     separator=True):
        """
        Writes a new prompt at the end of the buffer.
        """
        if prompt == self._pdb_prompt:
            html = True
            prompt = '<span class="in-prompt">%s</span>' % prompt
        super(DebuggingWidget, self)._show_prompt(prompt, html, newline,
                                                  separator)

    def _event_filter_console_keypress(self, event):
        """Handle Key_Up/Key_Down while debugging."""
        if self.is_waiting_pdb_input():
            self._control.current_prompt_pos = self._prompt_pos
            # Pretend this is a regular prompt
            self._tmp_reading = self._reading
            self._reading = False
            try:
                ret = super(DebuggingWidget,
                            self)._event_filter_console_keypress(event)
                return ret
            finally:
                self._reading = self._tmp_reading
        else:
            return super(DebuggingWidget,
                         self)._event_filter_console_keypress(event)

    def _register_is_complete_callback(self, source, callback):
        """Call the callback with the result of is_complete."""
        # Add a continuation prompt if not complete
        if self.is_waiting_pdb_input():
            # As the work is done on this side, check synchronously.
            complete, indent = self._is_pdb_complete(source)
            callback(complete, indent)
        else:
            return super(DebuggingWidget, self)._register_is_complete_callback(
                source, callback)

    # ---- Qt methods ---------------------------------------------------------
    def eventFilter(self, obj, event):
        # When using PySide, it can happen that "event" is of type QWidgetItem
        # (reason unknown). This causes an exception in eventFilter() in
        # console_widget.py in the QtConsole package: Therein event.type() is
        # accessed which fails due to an AttributeError. Catch this here and
        # ignore the event.
        if not isinstance(event, QEvent):
            # Note for debugging: event.layout() or event.widget() SEGFAULTs
            return True

        return super().eventFilter(obj, event)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Global plugin registry."""

# Standard library imports
import logging
from typing import Dict, List, Union, Type, Any, Set, Optional

# Third-party library imports
from qtpy.QtCore import QObject, Signal

# Local imports
from spyder import dependencies
from spyder.config.base import _, running_under_pytest
from spyder.config.manager import CONF
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.plugin_registration._confpage import PluginsConfigPage
from spyder.api.exceptions import SpyderAPIError
from spyder.api.plugins import Plugins, SpyderDockablePlugin, SpyderPluginV2
from spyder.api.plugins._old_api import SpyderPlugin, SpyderPluginWidget
from spyder.utils.icon_manager import ima


# TODO: Remove SpyderPlugin and SpyderPluginWidget now that the migration
# is complete
Spyder5PluginClass = Union[SpyderPluginV2, SpyderDockablePlugin]
Spyder4PluginClass = Union[SpyderPlugin, SpyderPluginWidget]
SpyderPluginClass = Union[Spyder4PluginClass, Spyder5PluginClass]


ALL_PLUGINS = [getattr(Plugins, attr) for attr in dir(Plugins)
               if not attr.startswith('_') and attr != 'All']

logger = logging.getLogger(__name__)


class PreferencesAdapter(SpyderConfigurationAccessor):
    # Fake class constants used to register the configuration page
    CONF_WIDGET_CLASS = PluginsConfigPage
    NAME = 'plugin_registry'
    CONF_VERSION = None
    ADDITIONAL_CONF_OPTIONS = None
    ADDITIONAL_CONF_TABS = None
    CONF_SECTION = ""

    def apply_plugin_settings(self, _unused):
        pass

    def apply_conf(self, _unused):
        pass


class SpyderPluginRegistry(QObject, PreferencesAdapter):
    """
    Global plugin registry.

    This class handles a plugin initialization/teardown lifetime, including
    notifications when a plugin is available or not.

    This registry alleviates the limitations of a topological sort-based
    plugin initialization by enabling plugins to have bidirectional
    dependencies instead of unidirectional ones.

    Notes
    -----
    1. This class should be instantiated as a singleton.
    2. A plugin should not depend on other plugin to perform its
       initialization since it could cause deadlocks.
    """

    sig_plugin_ready = Signal(str, bool)
    """
    This signal is used to let the main window know that a plugin is ready.

    Parameters
    ----------
    plugin_name: str
        Name of the plugin that is available.
    omit_conf: bool
        True if the plugin configuration does not need to be written.
    """

    def __init__(self):
        super().__init__()
        PreferencesAdapter.__init__(self)

        # Reference to the main window
        self.main = None

        # Dictionary that maps a plugin name to a list of the plugin names
        # that depend on it.
        self.plugin_dependents = {}  # type: Dict[str, Dict[str, List[str]]]

        # Dictionary that maps a plugin name to a list of the plugin names
        # that the plugin depends on.
        self.plugin_dependencies = {}  # type: Dict[str, Dict[str, List[str]]]

        # Plugin dictionary mapped by their names
        self.plugin_registry = {}  # type: Dict[str, SpyderPluginClass]

        # Dictionary that maps a plugin name to its availability.
        self.plugin_availability = {}  # type: Dict[str, bool]

        # Set that stores the plugin names of all Spyder 4 plugins.
        self.old_plugins = set({})  # type: set[str]

        # Set that stores the names of the plugins that are enabled
        self.enabled_plugins = set({})  # type: set[str]

        # Set that stores the names of the internal plugins
        self.internal_plugins = set({})  # type: set[str]

        # Set that stores the names of the external plugins
        self.external_plugins = set({})  # type: set[str]

        # Dictionary that contains all the internal plugins (enabled or not)
        self.all_internal_plugins = {}  # type: Dict[str, Tuple[str, Type[SpyderPluginClass]]]

        # Dictionary that contains all the external plugins (enabled or not)
        self.all_external_plugins = {}  # type: Dict[str, Tuple[str, Type[SpyderPluginClass]]]

        # This is used to allow disabling external plugins through Preferences
        self._external_plugins_conf_section = "external_plugins"

    # ------------------------- PRIVATE API -----------------------------------
    def _update_dependents(self, plugin: str, dependent_plugin: str, key: str):
        """Add `dependent_plugin` to the list of dependents of `plugin`."""
        plugin_dependents = self.plugin_dependents.get(plugin, {})
        plugin_strict_dependents = plugin_dependents.get(key, [])
        plugin_strict_dependents.append(dependent_plugin)
        plugin_dependents[key] = plugin_strict_dependents
        self.plugin_dependents[plugin] = plugin_dependents

    def _update_dependencies(self, plugin: str, required_plugin: str,
                             key: str):
        """Add `required_plugin` to the list of dependencies of `plugin`."""
        plugin_dependencies = self.plugin_dependencies.get(plugin, {})
        plugin_strict_dependencies = plugin_dependencies.get(key, [])
        plugin_strict_dependencies.append(required_plugin)
        plugin_dependencies[key] = plugin_strict_dependencies
        self.plugin_dependencies[plugin] = plugin_dependencies

    def _update_plugin_info(self, plugin_name: str,
                            required_plugins: List[str],
                            optional_plugins: List[str]):
        """Update the dependencies and dependents of `plugin_name`."""
        for plugin in required_plugins:
            self._update_dependencies(plugin_name, plugin, 'requires')
            self._update_dependents(plugin, plugin_name, 'requires')

        for plugin in optional_plugins:
            self._update_dependencies(plugin_name, plugin, 'optional')
            self._update_dependents(plugin, plugin_name, 'optional')

    def _instantiate_spyder5_plugin(
            self, main_window: Any,
            PluginClass: Type[Spyder5PluginClass],
            external: bool) -> Spyder5PluginClass:
        """Instantiate and register a Spyder 5+ plugin."""
        required_plugins = list(set(PluginClass.REQUIRES))
        optional_plugins = list(set(PluginClass.OPTIONAL))
        plugin_name = PluginClass.NAME

        logger.debug(f'Registering plugin {plugin_name} - {PluginClass}')

        if PluginClass.CONF_FILE:
            CONF.register_plugin(PluginClass)

        for plugin in list(required_plugins):
            if plugin == Plugins.All:
                required_plugins = list(set(required_plugins + ALL_PLUGINS))

        for plugin in list(optional_plugins):
            if plugin == Plugins.All:
                optional_plugins = list(set(optional_plugins + ALL_PLUGINS))

        # Update plugin dependency information
        self._update_plugin_info(plugin_name, required_plugins,
                                 optional_plugins)

        # Create and store plugin instance
        plugin_instance = PluginClass(main_window, configuration=CONF)
        self.plugin_registry[plugin_name] = plugin_instance

        # Connect plugin availability signal to notification system
        plugin_instance.sig_plugin_ready.connect(
            lambda: self.notify_plugin_availability(
                plugin_name, omit_conf=PluginClass.CONF_FILE))

        # Initialize plugin instance
        plugin_instance.initialize()

        # Register plugins that are already available
        self._notify_plugin_dependencies(plugin_name)

        # Register the plugin name under the external or internal
        # plugin set
        if external:
            self.external_plugins |= {plugin_name}
        else:
            self.internal_plugins |= {plugin_name}

        if external:
            # These attributes come from spyder.app.find_plugins
            module = PluginClass._spyder_module_name
            package_name = PluginClass._spyder_package_name
            version = PluginClass._spyder_version
            description = plugin_instance.get_description()
            dependencies.add(module, package_name, description,
                                version, None, kind=dependencies.PLUGIN)

        return plugin_instance

    def _instantiate_spyder4_plugin(
            self, main_window: Any,
            PluginClass: Type[Spyder4PluginClass],
            external: bool,
            args: tuple, kwargs: dict) -> Spyder4PluginClass:
        """Instantiate and register a Spyder 4 plugin."""
        plugin_name = PluginClass.NAME

        # Create old plugin instance
        plugin_instance = PluginClass(main_window, *args, **kwargs)

        if hasattr(plugin_instance, 'COMPLETION_PROVIDER_NAME'):
            if Plugins.Completions in self:
                completions = self.get_plugin(Plugins.Completions)
                completions.register_completion_plugin(plugin_instance)
        else:
            plugin_instance.register_plugin()

        # Register plugin in the registry
        self.plugin_registry[plugin_name] = plugin_instance

        # Register the name of the plugin under the external or
        # internal plugin set
        if external:
            self.external_plugins |= {plugin_name}
        else:
            self.internal_plugins |= {plugin_name}

        # Since Spyder 5+ plugins are loaded before old ones, preferences
        # will be available at this point.
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(plugin_instance)

        # Notify new-API plugins that depend on old ones
        self.notify_plugin_availability(plugin_name, False)

        return plugin_instance

    def _notify_plugin_dependencies(self, plugin_name: str):
        """Notify a plugin of its available dependencies."""
        plugin_instance = self.plugin_registry[plugin_name]
        plugin_dependencies = self.plugin_dependencies.get(plugin_name, {})
        required_plugins = plugin_dependencies.get('requires', [])
        optional_plugins = plugin_dependencies.get('optional', [])

        for plugin in required_plugins + optional_plugins:
            if plugin in self.plugin_registry:
                if self.plugin_availability.get(plugin, False):
                    logger.debug(f'Plugin {plugin} has already loaded')
                    plugin_instance._on_plugin_available(plugin)

    def _notify_plugin_teardown(self, plugin_name: str):
        """Notify dependents of a plugin that is going to be unavailable."""
        plugin_dependents = self.plugin_dependents.get(plugin_name, {})
        required_plugins = plugin_dependents.get('requires', [])
        optional_plugins = plugin_dependents.get('optional', [])

        for plugin in required_plugins + optional_plugins:
            if plugin in self.plugin_registry:
                if self.plugin_availability.get(plugin, False):
                    logger.debug(f'Notifying plugin {plugin} that '
                                 f'{plugin_name} is going to be turned off')
                    plugin_instance = self.plugin_registry[plugin]
                    plugin_instance._on_plugin_teardown(plugin_name)

    def _teardown_plugin(self, plugin_name: str):
        """Disconnect a plugin from its dependencies."""
        plugin_instance = self.plugin_registry[plugin_name]
        plugin_dependencies = self.plugin_dependencies.get(plugin_name, {})
        required_plugins = plugin_dependencies.get('requires', [])
        optional_plugins = plugin_dependencies.get('optional', [])

        for plugin in required_plugins + optional_plugins:
            if plugin in self.plugin_registry:
                if self.plugin_availability.get(plugin, False):
                    logger.debug(f'Disconnecting {plugin_name} from {plugin}')
                    plugin_instance._on_plugin_teardown(plugin)

    # -------------------------- PUBLIC API -----------------------------------
    def register_plugin(
            self, main_window: Any,
            PluginClass: Type[SpyderPluginClass],
            *args: tuple, external: bool = False,
            **kwargs: dict) -> SpyderPluginClass:
        """
        Register a plugin into the Spyder registry.

        Parameters
        ----------
        main_window: spyder.app.mainwindow.MainWindow
            Reference to Spyder's main window.
        PluginClass: type[SpyderPluginClass]
            The plugin class to register and create. It must be one of
            `spyder.app.registry.SpyderPluginClass`.
        *args: tuple
            Positional arguments used to initialize the plugin
            instance.
        external: bool
            If True, then the plugin is stored as a external plugin. Otherwise
            it will be marked as an internal plugin. Default: False
        **kwargs: dict
            Optional keyword arguments used to initialize the plugin instance.

        Returns
        -------
        plugin: SpyderPluginClass
            The instance of the registered plugin.

        Raises
        ------
        TypeError
            If the `PluginClass` does not inherit from any of
            `spyder.app.registry.SpyderPluginClass`

        Notes
        -----
        The optional `*args` and `**kwargs` will be removed once all
        plugins are migrated.
        """
        if not issubclass(PluginClass, (SpyderPluginV2, SpyderPlugin)):
            raise TypeError(f'{PluginClass} does not inherit from '
                            f'{SpyderPluginV2} nor {SpyderPlugin}')

        instance = None
        if issubclass(PluginClass, SpyderPluginV2):
            # Register a Spyder 5+ plugin
            instance = self._instantiate_spyder5_plugin(
                main_window, PluginClass, external)
        elif issubclass(PluginClass, SpyderPlugin):
            # Register a Spyder 4 plugin
            instance = self._instantiate_spyder4_plugin(
                main_window, PluginClass, external, args, kwargs)

        return instance

    def notify_plugin_availability(self, plugin_name: str,
                                   notify_main: bool = True,
                                   omit_conf: bool = False):
        """
        Notify dependent plugins of a given plugin of its availability.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin that is available.
        notify_main: bool
            If True, then a signal is emitted to the main window to perform
            further registration steps.
        omit_conf: bool
            If True, then the main window is instructed to not write the
            plugin configuration into the Spyder configuration file.
        """
        logger.debug(f'Plugin {plugin_name} has finished loading, '
                     'sending notifications')

        # Set plugin availability to True
        self.plugin_availability[plugin_name] = True

        # Notify the main window that the plugin is ready
        if notify_main:
            self.sig_plugin_ready.emit(plugin_name, omit_conf)

        # Notify plugin dependents
        plugin_dependents = self.plugin_dependents.get(plugin_name, {})
        required_plugins = plugin_dependents.get('requires', [])
        optional_plugins = plugin_dependents.get('optional', [])

        for plugin in required_plugins + optional_plugins:
            if plugin in self.plugin_registry:
                plugin_instance = self.plugin_registry[plugin]
                plugin_instance._on_plugin_available(plugin_name)

        if plugin_name == Plugins.Preferences and not running_under_pytest():
            plugin_instance = self.plugin_registry[plugin_name]
            plugin_instance.register_plugin_preferences(self)

    def can_delete_plugin(self, plugin_name: str) -> bool:
        """
        Check if a plugin from the registry can be deleted by its name.

        Paremeters
        ----------
        plugin_name: str
            Name of the plugin to check for deletion.

        Returns
        -------
        can_close: bool
            True if the plugin can be closed. False otherwise.
        """
        plugin_instance = self.plugin_registry[plugin_name]
        # Determine if plugin can be closed
        return plugin_instance.can_close()

    def dock_undocked_plugin(
            self, plugin_name: str, save_undocked: bool = False):
        """
        Dock plugin if undocked and save undocked state if requested

        Parameters
        ----------
        plugin_name: str
            Name of the plugin to check for deletion.
        save_undocked : bool, optional
            True if the undocked state needs to be saved. The default is False.

        Returns
        -------
        None.
        """
        plugin_instance = self.plugin_registry[plugin_name]

        if isinstance(plugin_instance, SpyderDockablePlugin):
            # Close undocked plugin if needed and save undocked state
            plugin_instance.close_window(save_undocked=save_undocked)
        elif isinstance(plugin_instance, SpyderPluginWidget):
            # Save if plugin was undocked to restore it the next time.
            if plugin_instance._undocked_window and save_undocked:
                plugin_instance.set_option(
                    'undocked_on_window_close', True)
            else:
                plugin_instance.set_option(
                    'undocked_on_window_close', False)

            # Close undocked plugins.
            plugin_instance._close_window()

    def delete_plugin(self, plugin_name: str, teardown: bool = True,
                      check_can_delete: bool = True) -> bool:
        """
        Remove and delete a plugin from the registry by its name.

        Paremeters
        ----------
        plugin_name: str
            Name of the plugin to delete.
        teardown: bool
            True if the teardown notification to other plugins should be sent
            when deleting the plugin, False otherwise.
        check_can_delete: bool
            True if the plugin should validate if it can be closed when this
            method is called, False otherwise.

        Returns
        -------
        plugin_deleted: bool
            True if the registry was able to teardown and remove the plugin.
            False otherwise.
        """
        logger.debug(f'Deleting plugin {plugin_name}')
        plugin_instance = self.plugin_registry[plugin_name]

        # Determine if plugin can be closed
        if check_can_delete:
            can_delete = self.can_delete_plugin(plugin_name)
            if not can_delete:
                return False

        if isinstance(plugin_instance, SpyderPluginV2):
            # Cleanly delete plugin widgets. This avoids segfautls with
            # PyQt 5.15
            if isinstance(plugin_instance, SpyderDockablePlugin):
                try:
                    plugin_instance.get_widget().close()
                    plugin_instance.get_widget().deleteLater()
                except RuntimeError:
                    pass
            else:
                container = plugin_instance.get_container()
                if container:
                    try:
                        container.close()
                        container.deleteLater()
                    except RuntimeError:
                        pass

            # Delete plugin
            try:
                plugin_instance.deleteLater()
            except RuntimeError:
                pass
            if teardown:
                # Disconnect plugin from other plugins
                self._teardown_plugin(plugin_name)

                # Disconnect depending plugins from the plugin to delete
                self._notify_plugin_teardown(plugin_name)

            # Perform plugin closure tasks
            try:
                plugin_instance.on_close(True)
            except RuntimeError:
                pass
        elif isinstance(plugin_instance, SpyderPlugin):
            try:
                plugin_instance.deleteLater()
            except RuntimeError:
                pass
            if teardown:
                # Disconnect depending plugins from the plugin to delete
                self._notify_plugin_teardown(plugin_name)

        # Delete plugin from the registry and auxiliary structures
        self.plugin_dependents.pop(plugin_name, None)
        self.plugin_dependencies.pop(plugin_name, None)
        if plugin_instance.CONF_FILE:
            # This must be done after on_close() so that plugins can modify
            # their (external) config therein.
            CONF.unregister_plugin(plugin_instance)

        for plugin in self.plugin_dependents:
            all_plugin_dependents = self.plugin_dependents[plugin]
            for key in {'requires', 'optional'}:
                plugin_dependents = all_plugin_dependents.get(key, [])
                if plugin_name in plugin_dependents:
                    plugin_dependents.remove(plugin_name)

        for plugin in self.plugin_dependencies:
            all_plugin_dependencies = self.plugin_dependencies[plugin]
            for key in {'requires', 'optional'}:
                plugin_dependencies = all_plugin_dependencies.get(key, [])
                if plugin_name in plugin_dependencies:
                    plugin_dependencies.remove(plugin_name)

        self.plugin_availability.pop(plugin_name)
        self.old_plugins -= {plugin_name}
        self.enabled_plugins -= {plugin_name}
        self.internal_plugins -= {plugin_name}
        self.external_plugins -= {plugin_name}

        # Remove the plugin from the registry
        self.plugin_registry.pop(plugin_name)

        return True

    def dock_all_undocked_plugins(self, save_undocked: bool = False):
        """
        Dock undocked plugins and save undocked state if required.

        Parameters
        ----------
        save_undocked : bool, optional
            True if the undocked state needs to be saved. The default is False.

        Returns
        -------
        None.

        """
        for plugin_name in (
                set(self.external_plugins) | set(self.internal_plugins)):
            self.dock_undocked_plugin(
                plugin_name, save_undocked=save_undocked)

    def can_delete_all_plugins(self,
                               excluding: Optional[Set[str]] = None) -> bool:
        """
        Determine if all plugins can be deleted except the ones to exclude.

        Parameters
        ----------
        excluding: Optional[Set[str]]
            A set that lists plugins (by name) that will not be deleted.

        Returns
        -------
        bool
            True if all plugins can be closed. False otherwise.
        """
        excluding = excluding or set({})
        can_close = True

        # Check external plugins
        for plugin_name in (
                set(self.external_plugins) | set(self.internal_plugins)):
            if plugin_name not in excluding:
                can_close &= self.can_delete_plugin(plugin_name)
                if not can_close:
                    break

        return can_close

    def delete_all_plugins(self, excluding: Optional[Set[str]] = None,
                           close_immediately: bool = False) -> bool:
        """
        Remove all plugins from the registry.

        The teardown mechanism will remove external plugins first and then
        internal ones, where the Spyder 4 plugins will be removed first and
        then the Spyder 5 ones.

        Parameters
        ----------
        excluding: Optional[Set[str]]
            A set that lists plugins (by name) that will not be deleted.
        close_immediately: bool
            If true, then the `can_close` status will be ignored.

        Returns
        -------
        all_deleted: bool
            True if all the plugins were closed and deleted. False otherwise.
        """
        excluding = excluding or set({})
        can_close = True

        # Check if all the plugins can be closed
        can_close = self.can_delete_all_plugins(excluding=excluding)

        if not can_close and not close_immediately:
            return False

        # Delete Spyder 4 external plugins
        for plugin_name in set(self.external_plugins):
            if plugin_name not in excluding:
                plugin_instance = self.plugin_registry[plugin_name]
                if isinstance(plugin_instance, SpyderPlugin):
                    can_close &= self.delete_plugin(
                        plugin_name, teardown=False, check_can_delete=False)
                    if not can_close and not close_immediately:
                        break

        if not can_close:
            return False

        # Delete Spyder 4 internal plugins
        for plugin_name in set(self.internal_plugins):
            if plugin_name not in excluding:
                plugin_instance = self.plugin_registry[plugin_name]
                if isinstance(plugin_instance, SpyderPlugin):
                    can_close &= self.delete_plugin(
                        plugin_name, teardown=False, check_can_delete=False)
                    if not can_close and not close_immediately:
                        break

        if not can_close:
            return False

        # Delete Spyder 5+ external plugins
        for plugin_name in set(self.external_plugins):
            if plugin_name not in excluding:
                plugin_instance = self.plugin_registry[plugin_name]
                if isinstance(plugin_instance, SpyderPluginV2):
                    can_close &= self.delete_plugin(
                        plugin_name, teardown=False, check_can_delete=False)
                    if not can_close and not close_immediately:
                        break

        if not can_close and not close_immediately:
            return False

        # Delete Spyder 5 internal plugins
        for plugin_name in set(self.internal_plugins):
            if plugin_name not in excluding:
                plugin_instance = self.plugin_registry[plugin_name]
                if isinstance(plugin_instance, SpyderPluginV2):
                    can_close &= self.delete_plugin(
                        plugin_name, teardown=False, check_can_delete=False)
                    if not can_close and not close_immediately:
                        break

        return can_close

    def get_plugin(self, plugin_name: str) -> SpyderPluginClass:
        """
        Get a reference to a plugin instance by its name.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin to retrieve.

        Returns
        -------
        plugin: SpyderPluginClass
            The instance of the requested plugin.

        Raises
        ------
        SpyderAPIError
            If the plugin name was not found in the registry.
        """
        if plugin_name in self.plugin_registry:
            plugin_instance = self.plugin_registry[plugin_name]
            return plugin_instance
        else:
            raise SpyderAPIError(f'Plugin {plugin_name} was not found in '
                                 'the registry')

    def set_plugin_enabled(self, plugin_name: str):
        """
        Add a plugin name to the set of enabled plugins.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin to add.
        """
        self.enabled_plugins |= {plugin_name}

    def is_plugin_enabled(self, plugin_name: str) -> bool:
        """
        Determine if a given plugin is enabled and is going to be
        loaded.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin to query.

        Returns
        -------
        plugin_enabled: bool
            True if the plugin is enabled and False if not.
        """
        return plugin_name in self.enabled_plugins

    def is_plugin_available(self, plugin_name: str) -> bool:
        """
        Determine if a given plugin was loaded and is available.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin to query.

        Returns
        -------
        plugin_available: bool
            True if the plugin is available and False if not.
        """
        return self.plugin_availability.get(plugin_name, False)

    def reset(self):
        """Reset and empty the plugin registry."""
        # Dictionary that maps a plugin name to a list of the plugin names
        # that depend on it.
        self.plugin_dependents = {}  # type: Dict[str, Dict[str, List[str]]]

        # Dictionary that maps a plugin name to a list of the plugin names
        # that the plugin depends on.
        self.plugin_dependencies = {}  # type: Dict[str, Dict[str, List[str]]]

        # Plugin dictionary mapped by their names
        self.plugin_registry = {}  # type: Dict[str, SpyderPluginClass]

        # Dictionary that maps a plugin name to its availability.
        self.plugin_availability = {}  # type: Dict[str, bool]

        # Set that stores the plugin names of all Spyder 4 plugins.
        self.old_plugins = set({})  # type: set[str]

        # Set that stores the names of the plugins that are enabled
        self.enabled_plugins = set({})

        # Set that stores the names of the internal plugins
        self.internal_plugins = set({})

        # Set that stores the names of the external plugins
        self.external_plugins = set({})

        try:
            self.sig_plugin_ready.disconnect()
        except (TypeError, RuntimeError):
            # Omit failures if there are no slots connected
            pass

        dependencies.DEPENDENCIES = []

    def set_all_internal_plugins(
            self, all_plugins: Dict[str, Type[SpyderPluginClass]]):
        self.all_internal_plugins = all_plugins

    def set_all_external_plugins(
            self, all_plugins: Dict[str, Type[SpyderPluginClass]]):
        self.all_external_plugins = all_plugins

    def set_main(self, main):
        self.main = main

    def get_icon(self):
        return ima.icon('plugins')

    def get_name(self):
        return _('Plugins')

    def __contains__(self, plugin_name: str) -> bool:
        """
        Determine if a plugin name is contained in the registry.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin to seek.

        Returns
        -------
        is_contained: bool
            If True, the plugin name is contained on the registry, False
            otherwise.
        """
        return plugin_name in self.plugin_registry

    def __iter__(self):
        return iter(self.plugin_registry)


PLUGIN_REGISTRY = SpyderPluginRegistry()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the widgets used in the Plots plugin.
"""

# Standard library imports
import os.path as osp
import datetime
from unittest.mock import Mock

# Third party imports
import pytest
from matplotlib.figure import Figure
import numpy as np
from qtpy.QtWidgets import QApplication, QStyle
from qtpy.QtGui import QPixmap
from qtpy.QtCore import Qt

# Local imports
from spyder.plugins.plots.widgets.figurebrowser import (FigureBrowser,
                                                        FigureThumbnail)
from spyder.plugins.plots.widgets.figurebrowser import get_unique_figname


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture
def figbrowser(qtbot):
    """An empty figure browser widget fixture."""
    figbrowser = FigureBrowser()
    figbrowser.set_shellwidget(Mock())
    options = {
        "mute_inline_plotting": True,
        "show_plot_outline": False,
    }
    figbrowser.setup(options)
    qtbot.addWidget(figbrowser)
    figbrowser.show()
    figbrowser.setMinimumSize(700, 500)
    return figbrowser


# =============================================================================
# ---- Helper functions
# =============================================================================
def create_figure(figname):
    """Create a matplotlib figure, save it to disk and return its data."""
    # Create and save to disk a figure with matplotlib.
    fig = Figure()
    ax = fig.add_axes([0.15, 0.15, 0.7, 0.7])
    fig.set_size_inches(6, 4)
    ax.plot(np.random.rand(10), '.', color='red')
    fig.savefig(figname)

    # Read back and return the binary data from the file.
    with open(figname, "rb") as img:
        fig = img.read()
    return fig


def add_figures_to_browser(figbrowser, nfig, tmpdir, fmt='image/png'):
    """
    Create and add bitmap figures to the figure browser. Also return a list
    of the created figures data.
    """
    fext = '.svg' if fmt == 'image/svg+xml' else '.png'
    figs = []
    for i in range(nfig):
        figname = osp.join(str(tmpdir), 'mplfig' + str(i) + fext)
        figs.append(create_figure(figname))
        figbrowser.add_figure(figs[-1], fmt)

    assert len(figbrowser.thumbnails_sb._thumbnails) == nfig
    assert figbrowser.thumbnails_sb.get_current_index() == nfig - 1
    assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == figs[-1]
    assert figbrowser.figviewer.figcanvas.fig == figs[-1]

    return figs


def png_to_qimage(png):
    """Return a QImage from the raw data of a png image."""
    qpix = QPixmap()
    qpix.loadFromData(png, 'image/png'.upper())
    return qpix.toImage()


# =============================================================================
# ---- Tests
# =============================================================================
@pytest.mark.order(1)
@pytest.mark.parametrize("fmt, fext",
                         [('image/png', '.png'), ('image/svg+xml', '.svg')])
def test_add_figures(figbrowser, tmpdir, fmt, fext):
    """
    Test that the figure browser widget display correctly new figures in
    its viewer and thumbnails scrollbar.
    """
    assert len(figbrowser.thumbnails_sb._thumbnails) == 0
    assert figbrowser.thumbnails_sb.current_thumbnail is None
    assert figbrowser.figviewer.figcanvas.fig is None

    for i in range(3):
        figname = osp.join(str(tmpdir), 'mplfig' + str(i) + fext)
        fig = create_figure(figname)
        figbrowser.add_figure(fig, fmt)
        assert len(figbrowser.thumbnails_sb._thumbnails) == i + 1
        assert figbrowser.thumbnails_sb.get_current_index() == i
        assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == fig
        assert figbrowser.figviewer.figcanvas.fig == fig


@pytest.mark.parametrize("fmt, fext",
                         [('image/png', '.png'),
                          ('image/svg+xml', '.svg'),
                          ('image/svg+xml', '.png')])
def test_save_figure_to_file(figbrowser, tmpdir, mocker, fmt, fext):
    """
    Test saving png and svg figures to file with the figure browser.
    """
    fig = add_figures_to_browser(figbrowser, 1, tmpdir, fmt)[0]
    expected_qpix = QPixmap()
    expected_qpix.loadFromData(fig, fmt.upper())

    # Save the figure to disk with the figure browser.
    saved_figname = osp.join(str(tmpdir), 'spyfig' + fext)
    mocker.patch('spyder.plugins.plots.widgets.figurebrowser.getsavefilename',
                 return_value=(saved_figname, fext))

    figbrowser.save_figure()
    saved_qpix = QPixmap()
    saved_qpix.load(saved_figname)

    assert osp.exists(saved_figname)
    assert expected_qpix.toImage() == saved_qpix.toImage()


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_save_all_figures(figbrowser, tmpdir, mocker, fmt):
    """
    Test saving all figures contained in the thumbnail scrollbar in batch
    into a single directory.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, fmt)

    # Save all figures, but cancel the dialog to get a directory.
    mocker.patch(
        'spyder.plugins.plots.widgets.figurebrowser.getexistingdirectory',
        return_value=None)
    fignames = figbrowser.save_all_figures()
    assert fignames is None

    # Save all figures.
    mocker.patch(
        'spyder.plugins.plots.widgets.figurebrowser.getexistingdirectory',
        return_value=str(tmpdir.mkdir('all_saved_figures')))
    fignames = figbrowser.save_all_figures()
    assert len(fignames) == len(figs)
    for fig, figname in zip(figs, fignames):
        expected_qpix = QPixmap()
        expected_qpix.loadFromData(fig, fmt.upper())
        saved_qpix = QPixmap()
        saved_qpix.load(figname)

        assert osp.exists(figname)
        assert expected_qpix.toImage() == saved_qpix.toImage()


def test_get_unique_figname(tmpdir):
    """
    Test that the unique fig names work when saving only one and when
    saving multiple figures.
    """
    fext = '.png'
    # Saving only one figure
    figname_root = ('Figure ' +
                    datetime.datetime.now().strftime('%Y-%m-%d %H%M%S'))
    figname = get_unique_figname(tmpdir, figname_root, fext)
    expected = osp.join(tmpdir, '{}{}'.format(figname_root, fext))
    assert figname == expected

    # Saving multiple figures
    figname_root = ('Figure ' +
                    datetime.datetime.now().strftime('%Y-%m-%d %H%M%S'))
    for i in range(5):
        figname = get_unique_figname(tmpdir, figname_root, fext,
                                     start_at_zero=True)
        # Create empty file with figname
        with open(figname, 'w') as _:
            pass

        expected = osp.join(tmpdir, '{} ({}){}'.format(figname_root, i, fext))
        assert figname == expected


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_close_current_figure(figbrowser, tmpdir, fmt):
    """
    Test that clearing the current figure works as expected.
    """
    figs = add_figures_to_browser(figbrowser, 2, tmpdir, fmt)

    # Remove the first figure.
    figbrowser.close_figure()
    assert len(figbrowser.thumbnails_sb._thumbnails) == 1
    assert figbrowser.thumbnails_sb.get_current_index() == 0
    assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == figs[0]
    assert figbrowser.figviewer.figcanvas.fig == figs[0]

    # Remove the last figure.
    figbrowser.close_figure()
    assert len(figbrowser.thumbnails_sb._thumbnails) == 0
    assert figbrowser.thumbnails_sb.get_current_index() == -1
    assert figbrowser.thumbnails_sb.current_thumbnail is None
    assert figbrowser.figviewer.figcanvas.fig is None


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_close_all_figures(figbrowser, tmpdir, fmt):
    """
    Test that clearing all figures displayed in the thumbnails scrollbar
    works as expected.
    """
    add_figures_to_browser(figbrowser, 3, tmpdir, fmt)

    # Close all previously opened figures.
    figbrowser.close_all_figures()
    assert len(figbrowser.thumbnails_sb._thumbnails) == 0
    assert figbrowser.thumbnails_sb.get_current_index() == -1
    assert figbrowser.thumbnails_sb.current_thumbnail is None
    assert figbrowser.figviewer.figcanvas.fig is None
    assert len(figbrowser.thumbnails_sb.findChildren(FigureThumbnail)) == 0


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_close_one_thumbnail(qtbot, figbrowser, tmpdir, fmt):
    """
    Test the thumbnail is removed from the GUI.
    """
    # Add two figures to the browser
    add_figures_to_browser(figbrowser, 2, tmpdir, fmt)
    assert len(figbrowser.thumbnails_sb.findChildren(FigureThumbnail)) == 2

    # Remove the first figure
    figures = figbrowser.thumbnails_sb.findChildren(FigureThumbnail)
    figbrowser.thumbnails_sb.remove_thumbnail(figures[0])

    # Removed thumbnails are unparented with a timer to prevent segfaults
    qtbot.wait(200)

    assert len(figbrowser.thumbnails_sb.findChildren(FigureThumbnail)) == 1


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_go_prev_next_thumbnail(figbrowser, tmpdir, fmt):
    """
    Test go to previous and next thumbnail actions.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, fmt)

    # Circle through the open figures with go_next_thumbnail and
    # go_previous_thumbnail.
    figbrowser.go_next_thumbnail()
    assert figbrowser.thumbnails_sb.get_current_index() == 0
    assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == figs[0]
    assert figbrowser.figviewer.figcanvas.fig == figs[0]

    figbrowser.go_previous_thumbnail()
    assert figbrowser.thumbnails_sb.get_current_index() == 2
    assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == figs[2]
    assert figbrowser.figviewer.figcanvas.fig == figs[2]

    figbrowser.go_previous_thumbnail()
    assert figbrowser.thumbnails_sb.get_current_index() == 1
    assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == figs[1]
    assert figbrowser.figviewer.figcanvas.fig == figs[1]


def test_scroll_to_item(figbrowser, tmpdir, qtbot):
    """Test scroll to the item of ThumbnailScrollBar."""
    nfig = 10
    add_figures_to_browser(figbrowser, nfig, tmpdir, 'image/png')
    figbrowser.setFixedSize(500, 500)

    for __ in range(nfig // 2):
        figbrowser.go_next_thumbnail()
        qtbot.wait(500)

    scene = figbrowser.thumbnails_sb.scene

    spacing = scene.spacing()
    height = scene.itemAt(0).sizeHint().height()
    height_view = figbrowser.thumbnails_sb.scrollarea.viewport().height()

    expected = (spacing * (nfig // 2)) + (height * (nfig // 2 - 1)) - \
               ((height_view - height) // 2)

    vsb = figbrowser.thumbnails_sb.scrollarea.verticalScrollBar()
    assert vsb.value() == expected


def test_scroll_down_to_newest_plot(figbrowser, tmpdir, qtbot):
    """
    Test that the ThumbnailScrollBar is scrolled to the newest plot after
    it is added to it.

    Test that covers spyder-ide/spyder#10914.
    """
    figbrowser.setFixedSize(500, 500)

    nfig = 8
    for i in range(8):
        newfig = create_figure(
            osp.join(str(tmpdir), 'new_mplfig{}.png'.format(i)))
        figbrowser.add_figure(newfig, 'image/png')
        qtbot.wait(500)

    # Assert that the scrollbar range was updated correctly and that it's
    # value was set to its maximum.
    height_view = figbrowser.thumbnails_sb.scrollarea.viewport().height()
    scene = figbrowser.thumbnails_sb.scene
    spacing = scene.spacing()
    height = scene.itemAt(0).sizeHint().height()

    expected = (spacing * (nfig - 1)) + (height * nfig) - height_view + 6
    vsb = figbrowser.thumbnails_sb.scrollarea.verticalScrollBar()
    assert vsb.value() == expected


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_mouse_clicking_thumbnails(figbrowser, tmpdir, qtbot, fmt):
    """
    Test mouse clicking on thumbnails.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, fmt)
    for i in [1, 0, 2]:
        qtbot.mouseClick(
            figbrowser.thumbnails_sb._thumbnails[i].canvas, Qt.LeftButton)
        assert figbrowser.thumbnails_sb.get_current_index() == i
        assert figbrowser.thumbnails_sb.current_thumbnail.canvas.fig == figs[i]
        assert figbrowser.figviewer.figcanvas.fig == figs[i]


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_save_thumbnails(figbrowser, tmpdir, qtbot, mocker, fmt):
    """
    Test saving figures by clicking on the thumbnail icon.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, fmt)
    fext = '.svg' if fmt == 'image/svg+xml' else '.png'

    # Select and save the second thumbnail of the scrollbar.
    figname = osp.join(str(tmpdir), 'figname' + fext)
    mocker.patch('spyder.plugins.plots.widgets.figurebrowser.getsavefilename',
                 return_value=(figname, fext))
    figbrowser.thumbnails_sb.set_current_index(1)
    figbrowser.save_figure()

    expected_qpix = QPixmap()
    expected_qpix.loadFromData(figs[1], fmt.upper())
    saved_qpix = QPixmap()
    saved_qpix.load(figname)

    assert osp.exists(figname)
    assert expected_qpix.toImage() == saved_qpix.toImage()


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_close_thumbnails(figbrowser, tmpdir, qtbot, mocker, fmt):
    """
    Test closing figures by clicking on the thumbnail icon.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, fmt)

    # Select and close the second thumbnail of the scrollbar.
    figbrowser.thumbnails_sb.set_current_index(1)
    figbrowser.close_figure()
    del figs[1]

    assert len(figbrowser.thumbnails_sb._thumbnails) == len(figs)
    assert figbrowser.thumbnails_sb._thumbnails[0].canvas.fig == figs[0]
    assert figbrowser.thumbnails_sb._thumbnails[1].canvas.fig == figs[1]


def test_copy_png_to_clipboard(figbrowser, tmpdir):
    """
    Test copying png figures to the clipboard.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, 'image/png')
    clipboard = QApplication.clipboard()

    # Copy the current figure (last thumbnail) to the clipboard.
    figbrowser.copy_figure()
    assert clipboard.image() == png_to_qimage(figs[-1])

    # Copy the first thumbnail to the clipboard.
    figbrowser.go_next_thumbnail()
    figbrowser.copy_figure()
    assert clipboard.image() == png_to_qimage(figs[0])


def test_copy_svg_to_clipboard(figbrowser, tmpdir):
    """
    Test copying svg figures to the clipboard.
    """
    figs = add_figures_to_browser(figbrowser, 3, tmpdir, 'image/svg+xml')
    clipboard = QApplication.clipboard()

    # Copy the current figure (last thumbnail) to the clipboard.
    figbrowser.copy_figure()
    assert clipboard.mimeData().data('image/svg+xml') == figs[-1]

    # Copy the first thumbnail to the clipboard.
    figbrowser.go_next_thumbnail()
    figbrowser.copy_figure()
    assert clipboard.mimeData().data('image/svg+xml') == figs[0]


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_zoom_figure_viewer(figbrowser, tmpdir, fmt):
    """
    Test zooming in and out the figure diplayed in the figure viewer.
    """
    fig = add_figures_to_browser(figbrowser, 1, tmpdir, fmt)[0]
    figcanvas = figbrowser.figviewer.figcanvas

    # Set `Fit plots to windows` to False before the test.
    figbrowser.figviewer.auto_fit_plotting = False

    # Calculate original figure size in pixels.
    qpix = QPixmap()
    qpix.loadFromData(fig, fmt.upper())
    fwidth, fheight = qpix.width(), qpix.height()

    assert figbrowser.zoom_disp_value < 100
    assert figcanvas.width() < fwidth
    assert figcanvas.height() < fheight

    # Zoom in and out the figure in the figure viewer.
    scaling_factor = figbrowser.figviewer.get_scale_factor()
    scaling_step = figbrowser.figviewer._scalestep
    for zoom_step in [1, 1, -1, -1, -1]:
        if zoom_step == 1:
            figbrowser.zoom_in()
        elif zoom_step == -1:
            figbrowser.zoom_out()

        scaling_factor += zoom_step
        scale = scaling_step ** scaling_factor

        assert (figbrowser.zoom_disp_value ==
                np.round(int(fwidth * scale) / fwidth * 100))
        assert figcanvas.width() == int(fwidth * scale)
        assert figcanvas.height() == int(fheight * scale)


@pytest.mark.parametrize("fmt", ['image/png', 'image/svg+xml'])
def test_autofit_figure_viewer(figbrowser, tmpdir, fmt):
    """
    Test figure diplayed when `Fit plots to window` is True.
    """
    fig = add_figures_to_browser(figbrowser, 1, tmpdir, fmt)[0]
    figviewer = figbrowser.figviewer
    figcanvas = figviewer.figcanvas

    # Calculate original figure size in pixels.
    qpix = QPixmap()
    qpix.loadFromData(fig, fmt.upper())
    fwidth, fheight = qpix.width(), qpix.height()

    # Test when `Fit plots to window` is set to True.
    # Otherwise, test should fall into `test_zoom_figure_viewer`
    figbrowser.figviewer.auto_fit_plotting = True

    size = figviewer.size()
    style = figviewer.style()
    width = (size.width() -
             style.pixelMetric(QStyle.PM_LayoutLeftMargin) -
             style.pixelMetric(QStyle.PM_LayoutRightMargin))
    height = (size.height() -
              style.pixelMetric(QStyle.PM_LayoutTopMargin) -
              style.pixelMetric(QStyle.PM_LayoutBottomMargin))
    if (fwidth / fheight) > (width / height):
        new_width = int(width)
        new_height = int(width / fwidth * fheight)
    else:
        new_height = int(height)
        new_width = int(height / fheight * fwidth)

    assert figcanvas.width() == new_width
    assert figcanvas.height() == new_height
    assert (figbrowser.zoom_disp_value ==
            round(figcanvas.width() / fwidth * 100))


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Helper classes to get and set the fonts used in Spyder.
"""

# Standard library imports
from typing import Optional

# Third-party imports
from qtpy.QtGui import QFont

# Local imports
from spyder.config.gui import get_font


class SpyderFontType:
    """
    Font types used in Spyder plugins and the entire application.

    Notes
    -----
    * This enum is meant to be used to get the QFont object corresponding to
      each type.
    * The names associated to the values in this enum depend on historical
      reasons that go back to Spyder 2 and are not easy to change now.
    * Monospace is the font used used in the Editor, IPython console and
      History; Interface is used by the entire Spyder app; and
      MonospaceInterface is used, for instance, by the Variable Explorer and
      corresponds to Monospace font resized to look good against the
      Interface one.
    """
    Monospace = 'font'
    Interface = 'app_font'
    MonospaceInterface = 'monospace_app_font'


class SpyderFontsMixin:
    """Mixin to get the different Spyder font types from our config system."""

    @classmethod
    def get_font(
        cls,
        font_type: str,
        font_size_delta: Optional[int] = 0
    ) -> QFont:
        """
        Get a font type as a QFont object.

        Parameters
        ----------
        font_type: str
            A Spyder font type. This must be one of the `SpyderFontType` enum
            values.
        font_size_delta: int, optional
            Small increase or decrease over the default font size. The default
            is 0.
        """
        return get_font(option=font_type, font_size_delta=font_size_delta)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Status bar widgets."""

# Standard library imports
import os.path as osp
import datetime

# Third-party imports
from qtpy.QtCore import Qt, QUrl
from qtpy.QtWidgets import QDialog, QVBoxLayout

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.widgets.status import BaseTimerStatus
from spyder.api.translations import _
from spyder.config.base import get_module_source_path
from spyder.config.gui import is_dark_interface
from spyder.utils.icon_manager import ima
from spyder.utils.qthelpers import start_file
from spyder.utils.stylesheet import WIN
from spyder.widgets.browser import WebView


class InAppAppealDialog(QDialog, SpyderFontsMixin):

    CONF_SECTION = "main"
    WIDTH = 530
    HEIGHT = 620 if WIN else 665

    def __init__(self, parent=None):
        super().__init__(parent)

        # Attributes
        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint
        )
        self.setFixedWidth(self.WIDTH)
        self.setFixedHeight(self.HEIGHT)
        self.setWindowTitle(_("Help Spyder"))
        self.setWindowIcon(ima.icon("inapp_appeal"))

        # Path to the appeal page
        appeal_page_path = osp.join(
            get_module_source_path("spyder.plugins.application.widgets"),
            "appeal_page",
            "dark" if is_dark_interface() else "light",
            "index.html",
        )

        # Create webview to render the appeal message
        webview = WebView(self, handle_links=True)

        # Set font used in the view
        app_font = self.get_font(SpyderFontType.Interface)
        webview.set_font(app_font, size_delta=2)

        # Load page
        webview.load(QUrl.fromLocalFile(appeal_page_path))

        # Open links in external browser
        webview.page().linkClicked.connect(self._handle_link_clicks)

        # Layout
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(webview)
        self.setLayout(layout)

    def _handle_link_clicks(self, url):
        url = str(url.toString())
        if url.startswith('http'):
            start_file(url)


class InAppAppealStatus(BaseTimerStatus):
    """Status bar widget for current file read/write mode."""

    ID = "inapp_appeal_status"
    CONF_SECTION = "main"
    INTERACT_ON_CLICK = True

    DAYS_TO_SHOW_AGAIN = 15

    def __init__(self, parent=None):
        super().__init__(parent)

        self._is_shown = False
        self._appeal_dialog = None

        # We don't need to show a label for this widget
        self.label_value.setVisible(False)

        # Update status every hour
        self.set_interval(60 * 60 * 1000)

        # Show appeal on click
        self.sig_clicked.connect(self._on_click)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _on_click(self):
        """Handle widget clicks."""
        if self._appeal_dialog is None:
            self._appeal_dialog = InAppAppealDialog(self)

        if self._appeal_dialog.isVisible():
            self._appeal_dialog.hide()
        else:
            self._appeal_dialog.show()

    # ---- Public API
    # -------------------------------------------------------------------------
    def show_appeal(self):
        if self._appeal_dialog is None:
            self._appeal_dialog = InAppAppealDialog(self)

        if not self._appeal_dialog.isVisible():
            self._appeal_dialog.show()

    # ---- StatusBarWidget API
    # -------------------------------------------------------------------------
    def get_icon(self):
        return self.create_icon("inapp_appeal")

    def update_status(self):
        """
        Show widget for a day after a certain number of days, then hide it.
        """
        today = datetime.date.today()
        last_date = self.get_conf("last_inapp_appeal", default="")

        if last_date:
            delta = today - datetime.date.fromisoformat(last_date)
            if 0 < delta.days < self.DAYS_TO_SHOW_AGAIN:
                self.setVisible(False)
            else:
                self.setVisible(True)
                self.set_conf("last_inapp_appeal", str(today))
        else:
            self.set_conf("last_inapp_appeal", str(today))

    def get_tooltip(self):
        return _("Help Spyder!")

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event):
        super().showEvent(event)

        # Hide widget if necessary at startup
        if not self._is_shown:
            self.update_status()
            self._is_shown = True



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from __future__ import annotations
import asyncio
import asyncio.events
from asyncio.tasks import _current_tasks
import atexit
from concurrent.futures import CancelledError, Future
import contextlib
from contextlib import contextmanager
import functools
from heapq import heappop
import logging
import os
import sys
import threading
import typing


if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec  # noqa: ICN003

_logger = logging.getLogger(__name__)


LoopID = typing.Union[typing.Hashable, asyncio.AbstractEventLoop]

P = ParamSpec("P")

T = typing.TypeVar("T")


class AsyncDispatcher:
    """Decorator to convert a coroutine to a sync function.

    Helper class to facilitate the conversion of coroutines to sync functions
    or to run a coroutine as a sync function without the need to call the event
    loop method.

    Usage
    ------
    As a decorator:
    ```
    @AsyncDispatcher.dispatch()
    async def my_coroutine():
        pass

    my_coroutine()
    ```

    As a class wrapper:
    ```
    sync_coroutine = AsyncDispatcher(my_coroutine)

    sync_coroutine()
    ```
    """

    __rlock = threading.RLock()

    __closed = False
    __running_threads: typing.ClassVar[dict[typing.Hashable, _LoopRunner]] = {}
    _running_tasks: typing.ClassVar[list[Future]] = []

    def __init__(self,
                 async_func: typing.Callable[..., typing.Coroutine[
                             typing.Any, typing.Any, typing.Any]],
                 *,
                 loop: LoopID | None = None,
                 early_return: bool = True):
        """Initialize the decorator.

        Parameters
        ----------
        coro : coroutine
            The coroutine to be wrapped.
        loop : asyncio.AbstractEventLoop, optional
            The event loop to be used, by default get the current event loop.
        """
        if not asyncio.iscoroutinefunction(async_func):
            msg = f"{async_func} is not a coroutine function"
            raise TypeError(msg)
        self._async_func = async_func
        self._loop = self._ensure_running_loop(loop)
        self._early_return = early_return

    def __call__(self, *args, **kwargs):
        task = asyncio.run_coroutine_threadsafe(
            self._async_func(*args, **kwargs), loop=self._loop
        )
        if self._early_return:
            AsyncDispatcher._running_tasks.append(task)
            task.add_done_callback(self._callback_task_done)
            return task
        return task.result()

    @classmethod
    def dispatch(cls,
                 *,
                 loop: LoopID | None = None,
                 early_return: bool = True):
        """Create a decorator to run the coroutine with a given event loop."""

        def decorator(
            async_func: typing.Callable[P, typing.Coroutine[typing.Any,
                                                            typing.Any, T]]
        ) -> typing.Callable[P, Future[T] | T]:
            @functools.wraps(async_func)
            def wrapper(*args, **kwargs):
                return cls(async_func, loop=loop, early_return=early_return)(
                    *args, **kwargs
                )

            return wrapper

        return decorator

    def _callback_task_done(self, future: Future):
        AsyncDispatcher._running_tasks.remove(future)
        with contextlib.suppress(asyncio.CancelledError, CancelledError):
            if (exception := future.exception()) is not None:
                raise exception

    @classmethod
    def _ensure_running_loop(
        cls,
        loop_id: LoopID | None = None
    ) -> asyncio.AbstractEventLoop:
        loop, loop_id = cls.get_event_loop(loop_id)

        try:
            if loop.is_running():
                return loop
        except RuntimeError:
            _logger.exception(
                "Failed to check if the loop is running, defaulting to the "
                "current loop."
            )
            return asyncio.get_event_loop()

        with cls.__rlock:
            # Re-check, perhaps it was created in the meantime...
            if loop_id not in cls.__running_threads:
                cls.__run_loop(loop_id, loop)
                if loop_id is None:
                    asyncio.set_event_loop(loop)

        return loop

    @classmethod
    def get_event_loop(
        cls, loop_id: LoopID | None = None
    ) -> tuple[asyncio.AbstractEventLoop, typing.Hashable | None]:
        if loop_id is None:
            try:
                return asyncio.get_running_loop(), None
            except RuntimeError:  # noqa: S110
                pass
        elif isinstance(loop_id, asyncio.AbstractEventLoop):
            return loop_id, hash(loop_id)

        running_thread = cls.__running_threads.get(loop_id)
        if running_thread is not None:
            return running_thread.loop, loop_id

        return asyncio.new_event_loop(), loop_id

    @classmethod
    def __run_loop(cls,
                   loop_id: typing.Hashable,
                   loop: asyncio.AbstractEventLoop):
        if loop_id not in cls.__running_threads:
            with cls.__rlock:
                if loop_id not in cls.__running_threads:
                    _patch_loop_as_reentrant(loop)  # ipykernel compatibility
                    cls.__running_threads[loop_id] = \
                        _LoopRunner(loop_id, loop)
                    cls.__running_threads[loop_id].start()

    @staticmethod
    @atexit.register
    def close():
        """Close the thread pool."""
        if AsyncDispatcher.__closed:
            return
        AsyncDispatcher.cancel_all()
        AsyncDispatcher.join()
        AsyncDispatcher.__closed = True

    @classmethod
    def cancel_all(cls):
        """Cancel all running tasks."""
        for task in cls._running_tasks:
            task.cancel()

    @classmethod
    def join(cls, timeout: float | None = None):
        """Close all running loops and join the threads."""
        for loop_id in list(cls.__running_threads.keys()):
            cls._stop_running_loop(loop_id, timeout)

    @classmethod
    def _stop_running_loop(cls,
                           loop_id: LoopID,
                           timeout: int | None = None):
        runner = cls.__running_threads.pop(loop_id, None)

        if runner is None:
            return

        runner.join(timeout)


class _LoopRunner(threading.Thread):
    """A task runner that runs an asyncio event loop on a background thread."""

    def __init__(self, loop_id: str,
                 loop: asyncio.AbstractEventLoop):
        super().__init__(daemon=True, name=f"AsyncDispatcher-{loop_id}")
        self.__loop = loop
        self.__loop_stopped = threading.Event()

    @property
    def loop(self):
        return self.__loop

    def run(self):
        asyncio.set_event_loop(self.__loop)
        try:
            self.__loop.run_forever()
        finally:
            self.__loop.close()
            self.__loop_stopped.set()

    def stop(self):
        self.__loop.call_soon_threadsafe(self.__loop.stop)

    def join(self, timeout: float | None = None):
        if not self.__loop_stopped.is_set():
            self.stop()
            self.__loop_stopped.wait(timeout)
        super().join(timeout)


def _patch_loop_as_reentrant(loop):
    """Patch an event loop in order to make it reentrant.

    This is a simplified version of the 'nest_asyncio'.
    """

    if hasattr(loop, '_nest_patched'):
        """Use same check as asyncio to avoid re-patching."""
        return

    def run_forever(self):
        with manage_run(self), manage_asyncgens(self):
            while True:
                self._run_once()
                if self._stopping:
                    break
        self._stopping = False

    def run_until_complete(self, future):
        with manage_run(self):
            f = asyncio.ensure_future(future, loop=self)
            if f is not future:
                f._log_destroy_pending = False
            while not f.done():
                self._run_once()
                if self._stopping:
                    break
            if not f.done():
                raise RuntimeError(
                    'Event loop stopped before Future completed.')
            return f.result()

    def _run_once(self):
        """
        Simplified re-implementation of asyncio's _run_once that
        runs handles as they become ready.
        """
        ready = self._ready
        scheduled = self._scheduled
        while scheduled and scheduled[0]._cancelled:
            heappop(scheduled)

        timeout = (
            0 if ready or self._stopping
            else min(max(
                scheduled[0]._when - self.time(), 0), 86400) if scheduled
            else None)
        event_list = self._selector.select(timeout)
        self._process_events(event_list)

        end_time = self.time() + self._clock_resolution
        while scheduled and scheduled[0]._when < end_time:
            handle = heappop(scheduled)
            ready.append(handle)

        for _ in range(len(ready)):
            if not ready:
                break
            handle = ready.popleft()
            if not handle._cancelled:
                # preempt the current task so that that checks in
                # Task.__step do not raise
                curr_task = _current_tasks.pop(self, None)

                try:
                    handle._run()
                finally:
                    # restore the current task
                    if curr_task is not None:
                        _current_tasks[self] = curr_task

        handle = None

    @contextmanager
    def manage_run(self):
        """Set up the loop for running."""
        self._check_closed()
        old_thread_id = self._thread_id
        old_running_loop = asyncio.events._get_running_loop()
        try:
            self._thread_id = threading.get_ident()
            asyncio.events._set_running_loop(self)
            self._num_runs_pending += 1
            if self._is_proactorloop:
                if self._self_reading_future is None:
                    self.call_soon(self._loop_self_reading)
            yield
        finally:
            self._thread_id = old_thread_id
            asyncio.events._set_running_loop(old_running_loop)
            self._num_runs_pending -= 1
            if self._is_proactorloop:
                if (self._num_runs_pending == 0
                        and self._self_reading_future is not None):
                    ov = self._self_reading_future._ov
                    self._self_reading_future.cancel()
                    if ov is not None:
                        self._proactor._unregister(ov)
                    self._self_reading_future = None

    @contextmanager
    def manage_asyncgens(self):
        if not hasattr(sys, 'get_asyncgen_hooks'):
            # Python version is too old.
            return
        old_agen_hooks = sys.get_asyncgen_hooks()
        try:
            self._set_coroutine_origin_tracking(self._debug)
            if self._asyncgens is not None:
                sys.set_asyncgen_hooks(
                    firstiter=self._asyncgen_firstiter_hook,
                    finalizer=self._asyncgen_finalizer_hook)
            yield
        finally:
            self._set_coroutine_origin_tracking(False)
            if self._asyncgens is not None:
                sys.set_asyncgen_hooks(*old_agen_hooks)

    def _check_running(self):
        """Do not throw exception if loop is already running."""
        pass

    if not isinstance(loop, asyncio.BaseEventLoop):
        raise ValueError('Can\'t patch loop of type %s' % type(loop))
    cls = loop.__class__
    cls.run_forever = run_forever
    cls.run_until_complete = run_until_complete
    cls._run_once = _run_once
    cls._check_running = _check_running
    cls._num_runs_pending = 1 if loop.is_running() else 0
    cls._is_proactorloop = (
        os.name == 'nt' and issubclass(cls, asyncio.ProactorEventLoop))
    cls._nest_patched = True



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for code completion."""

# Standard library imports
import os
import os.path as osp
import random
import textwrap
import sys

# Third party imports
from flaky import flaky
import pytest
import pytestqt
from qtpy.QtCore import Qt
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.config.utils import is_anaconda
from spyder.plugins.completion.api import (
    CompletionRequestTypes, CompletionItemKind)
from spyder.plugins.completion.providers.languageserver.providers.utils import (
    path_as_uri)
from spyder.utils.conda import get_list_conda_envs


# Location of this file
LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))


def set_executable_helper(completion_plugin, executable=None):
    if executable is None:
        completion_plugin._sig_interpreter_changed.emit(sys.executable)
    else:
        completion_plugin._sig_interpreter_changed.emit(executable)


@pytest.mark.order(1)
@pytest.mark.skipif(not sys.platform.startswith('linux'),
                    reason='Only works on Linux')
@flaky(max_runs=5)
def test_fallback_completions(completions_codeeditor, qtbot):
    code_editor, completion_plugin = completions_codeeditor
    completion = code_editor.completion_widget

    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    # Set cursor to start
    code_editor.go_to_line(1)

    # Add some words in comments
    qtbot.keyClicks(code_editor, '# some comment and whole words')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=1000)

    qtbot.keyClicks(code_editor, 'wh')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    # Assert all retrieved words start with 'wh'
    assert all({x['insertText'].startswith('wh') for x in sig.args[0]})

    # Delete 'wh'
    for _ in range(2):
        qtbot.keyPress(code_editor, Qt.Key_Backspace)

    # Insert another word
    qtbot.keyClicks(code_editor, 'another')

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    qtbot.keyClicks(code_editor, 'a')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    word_set = {x['insertText'] for x in sig.args[0]}
    assert 'another' in word_set

    # Assert that keywords are also retrieved
    assert 'assert' in word_set

    for _ in range(3):
        qtbot.keyPress(code_editor, Qt.Key_Backspace)

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, 'a')
        qtbot.keyPress(code_editor, Qt.Key_Tab, delay=300)

    word_set = {x['insertText'] for x in sig.args[0]}
    assert 'another' not in word_set

    # Check that fallback doesn't give an error with utf-16 characters.
    # This is a regression test for issue spyder-ide/spyder#11862.
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        code_editor.append("'😒 foobar'")
        qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
        qtbot.keyClicks(code_editor, 'foob')
        qtbot.keyPress(code_editor, Qt.Key_Tab, delay=300)

    word_set = {x['insertText'] for x in sig.args[0]}
    assert 'foobar' in word_set

    code_editor.toggle_automatic_completions(True)
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
def test_space_completion(completions_codeeditor, qtbot):
    """Validate completion's space character handling."""
    code_editor, _ = completions_codeeditor

    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    completion = code_editor.completion_widget

    # Set cursor to start
    code_editor.go_to_line(1)

    # Complete from numpy --> from numpy import
    qtbot.keyClicks(code_editor, 'from numpy ')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # press tab and get completions
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(completion, Qt.Key_Tab)

    assert "import" in [x['label'] for x in sig.args[0]]

    assert code_editor.toPlainText() == 'from numpy import'
    assert not completion.isVisible()

    code_editor.toggle_automatic_completions(True)
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@flaky(max_runs=5)
@pytest.mark.skipif(running_in_ci(), reason='Fails on CI!')
def test_hide_widget_completion(completions_codeeditor, qtbot):
    """Validate hiding completion widget after a delimeter or operator."""
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget

    delimiters = ['(', ')', '[', ']', '{', '}', ',', ':', ';', '@', '=', '->',
                  '+=', '-=', '*=', '/=', '//=', '%=', '@=', '&=', '|=', '^=',
                  '>>=', '<<=', '**=']

    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    # Set cursor to start
    code_editor.set_text('')
    code_editor.completion_widget.hide()
    code_editor.go_to_line(1)

    # Complete from numpy import --> from numpy import ?
    qtbot.keyClicks(code_editor, 'from numpy import ')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # Press tab and get completions
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000):
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    # Check the completion widget is visible
    assert completion.isHidden() is False

    # Write a random delimeter on the code editor
    delimiter = random.choice(delimiters)
    print(delimiter)
    qtbot.keyClicks(code_editor, delimiter)
    qtbot.wait(1000)

    # Check the completion widget is not visible
    assert completion.isHidden() is True

    code_editor.toggle_automatic_completions(True)
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@flaky(max_runs=10)
def test_automatic_completions(completions_codeeditor, qtbot):
    """Test on-the-fly completions."""
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    delay = 50
    code_editor.toggle_code_snippets(False)

    # Set cursor to start
    code_editor.go_to_line(1)

    # Complete f -> from
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=3000) as sig:
        qtbot.keyClicks(code_editor, 'f', delay=delay)

    assert "from" in [x['label'] for x in sig.args[0]]
    # qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyClicks(code_editor, 'rom')

    # Due to automatic completion, the completion widget may appear before
    stop = False
    while not stop:
        try:
            with qtbot.waitSignal(completion.sig_show_completions,
                                  timeout=5000) as sig:
                pass
            code_editor.completion_widget.hide()
        except Exception:
            stop = True

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, ' n', delay=delay)

    assert "ntpath" in [x['label'] for x in sig.args[0]]

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, 'ump', delay=delay)

    assert "numpy" in [x['label'] for x in sig.args[0]]

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, 'y', delay=delay)

    # Due to automatic completion, the completion widget may appear before
    stop = False
    while not stop:
        try:
            with qtbot.waitSignal(completion.sig_show_completions,
                                  timeout=5000) as sig:
                pass
            code_editor.completion_widget.hide()
        except Exception:
            stop = True

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, ' imp', delay=delay)

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(completion, Qt.Key_Tab)

    assert code_editor.toPlainText() == 'from numpy import'

    # Due to automatic completion, the completion widget may appear before
    stop = False
    while not stop:
        try:
            with qtbot.waitSignal(completion.sig_show_completions,
                                  timeout=5000) as sig:
                pass
            code_editor.completion_widget.hide()
        except Exception:
            stop = True

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, ' r', delay=delay)

    assert "random" in [x['label'] for x in sig.args[0]]
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_automatic_completions_tab_bug(completions_codeeditor, qtbot):
    """
    Test on-the-fly completions.

    Autocompletions should not be invoked when Tab/Backtab is pressed.

    See: spyder-ide/spyder#11625
    """
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    code_editor.toggle_code_snippets(False)

    code_editor.set_text('x = 1')
    code_editor.set_cursor_position('sol')

    try:
        with qtbot.waitSignal(completion.sig_show_completions,
                              timeout=5000):
            qtbot.keyPress(code_editor, Qt.Key_Tab)
        assert False
    except pytestqt.exceptions.TimeoutError:
        pass

    try:
        with qtbot.waitSignal(completion.sig_show_completions,
                              timeout=5000):
            qtbot.keyPress(code_editor, Qt.Key_Backtab)
        assert False
    except pytestqt.exceptions.TimeoutError:
        pass


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_automatic_completions_space_bug(completions_codeeditor, qtbot):
    """Test that completions are not invoked when pressing the space key."""
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    code_editor.toggle_code_snippets(False)

    code_editor.set_text('x = 1')
    code_editor.set_cursor_position('sol')
    qtbot.keyPress(code_editor, Qt.Key_Right)

    try:
        with qtbot.waitSignal(completion.sig_show_completions,
                              timeout=5000):
            qtbot.keyPress(code_editor, Qt.Key_Space)
        assert False
    except pytestqt.exceptions.TimeoutError:
        pass


@flaky(max_runs=10)
def test_automatic_completions_parens_bug(completions_codeeditor, qtbot):
    """
    Test on-the-fly completions.

    Autocompletions for variables don't work inside function calls.
    Note: Don't mark this as first because it fails on Windows.

    See: spyder-ide/spyder#10448
    """
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    delay = 50
    code_editor.toggle_code_snippets(False)

    # Parens:
    # Set cursor to start
    code_editor.set_text('my_list = [1, 2, 3]\nlist_copy = list((my))')
    qtbot.wait(500)
    cursor = code_editor.textCursor()
    code_editor.moveCursor(cursor.End)

    # Move cursor next to list((my$))
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.wait(500)

    # Complete my_ -> my_list
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=5000) as sig:
        qtbot.keyClicks(code_editor, '_', delay=delay)

    assert "my_list" in [x['label'] for x in sig.args[0]]
    qtbot.keyPress(completion, Qt.Key_Enter)

    # Square braces:
    # Set cursor to start
    code_editor.set_text('my_dic = {1: 1, 2: 2}\nonesee = 1\none = my_dic[on]')
    qtbot.wait(500)
    cursor = code_editor.textCursor()
    code_editor.moveCursor(cursor.End)

    # Move cursor next to my_dic[on$]
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.wait(500)

    # Complete one -> onesee
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=5000) as sig:
        qtbot.keyClicks(code_editor, 'e', delay=delay)

    assert "onesee" in [x['label'] for x in sig.args[0]]
    qtbot.keyPress(completion, Qt.Key_Enter)

    # Curly braces:
    # Set cursor to start
    code_editor.set_text('my_dic = {1: 1, 2: 2}\nonesee = 1\none = {on}')
    qtbot.wait(500)
    cursor = code_editor.textCursor()
    code_editor.moveCursor(cursor.End)

    # Move cursor next to {on*}
    qtbot.keyPress(code_editor, Qt.Key_Left)
    qtbot.wait(500)

    # Complete one -> onesee
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=5000) as sig:
        qtbot.keyClicks(code_editor, 'e', delay=delay)

    assert "onesee" in [x['label'] for x in sig.args[0]]
    qtbot.keyPress(completion, Qt.Key_Enter)


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_completions(completions_codeeditor, qtbot):
    """Exercise code completion in several ways."""
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget

    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    # Set cursor to start
    code_editor.go_to_line(1)

    # Complete dunder imports from _ --> import _foo/_foom
    qtbot.keyClicks(code_editor, 'from _')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # press tab and get completions
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert "__future__" in [x['label'] for x in sig.args[0]]
    code_editor.set_text('')  # Delete line
    code_editor.go_to_line(1)

    # Complete underscore variables
    qtbot.keyClicks(code_editor, '_foo = 1;_foom = 2;_fo')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # press tab and get completions
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    completions = [x['label'] for x in sig.args[0]]
    assert "_foo" in completions
    assert "_foom" in completions
    code_editor.set_text('')  # Delete line
    code_editor.go_to_line(1)

    # Complete import mat--> import math
    qtbot.keyClicks(code_editor, 'import mat')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    # press tab and get completions
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert "math" in [x['label'] for x in sig.args[0]]

    # enter should accept first completion
    qtbot.keyPress(completion, Qt.Key_Enter, delay=300)
    assert code_editor.toPlainText() == 'import math'

    # enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)

    # Complete math.h() -> math.hypot()
    qtbot.keyClicks(code_editor, 'math.h')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(completion, Qt.Key_Tab)

    assert [x['label'] for x in sig.args[0]][0] in ["hypot(x, y)",
                                                    "hypot(*coordinates)",
                                                    'hypot(coordinates)']

    print([(x['label'], x['provider']) for x in sig.args[0]])

    assert code_editor.toPlainText() == 'import math\nmath.hypot'

    qtbot.keyPress(code_editor, Qt.Key_Escape)

    try:
        with qtbot.waitSignal(completion.sig_show_completions,
                              timeout=10000) as sig:
            qtbot.keyPress(code_editor, Qt.Key_Tab)
    except pytestqt.exceptions.TimeoutError:
        # This should generate a timeout error because the completion
        # prefix is the same that the completions returned by Jedi.
        # This is a regression test for spyder-ide/spyder#11600
        pass

    # enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)

    # Complete math.h() -> math.degrees()
    qtbot.keyClicks(code_editor, 'math.h(')
    qtbot.keyPress(code_editor, Qt.Key_Left, delay=300)
    qtbot.keyClicks(code_editor, 'y')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert [x['label'] for x in sig.args[0]][0] in ["hypot(x, y)",
                                                    "hypot(*coordinates)",
                                                    'hypot(coordinates)']

    # right for () + enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Right, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Right, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    assert code_editor.toPlainText() == 'import math\nmath.hypot\n'\
                                        'math.hypot()\n'

    # Complete math.a <tab> ... s <enter> to math.asin
    qtbot.keyClicks(code_editor, 'math.a')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert "asin(x)" in [x['label'] for x in sig.args[0]]
    # Test if the list is updated
    assert "acos(x)" == completion.completion_list[0]['label']
    qtbot.keyClicks(completion, 's')
    data = completion.item(0).data(Qt.UserRole)
    assert "asin" == data['insertText']
    qtbot.keyPress(completion, Qt.Key_Enter, delay=300)

    # enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    assert code_editor.toPlainText() == 'import math\nmath.hypot\n'\
                                        'math.hypot()\nmath.asin\n'

    # Check can get list back
    qtbot.keyClicks(code_editor, 'math.f')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert completion.count() == 6
    assert "floor(x)" in [x['label'] for x in sig.args[0]]
    qtbot.keyClicks(completion, 'l')
    assert completion.count() == 1
    qtbot.keyPress(completion, Qt.Key_Backspace)
    assert completion.count() == 6

    # enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    assert code_editor.toPlainText() == 'import math\nmath.hypot\n'\
                                        'math.hypot()\nmath.asin\n'\
                                        'math.f\n'

    # Complete math.a <tab> s ...<enter> to math.asin
    qtbot.keyClicks(code_editor, 'math.a')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
        qtbot.keyPress(code_editor, 's')
    assert "asin(x)" in [x['label'] for x in sig.args[0]]
    qtbot.keyPress(completion, Qt.Key_Enter, delay=300)

    # enter for new line
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    assert code_editor.toPlainText() == 'import math\nmath.hypot\n'\
                                        'math.hypot()\nmath.asin\n'\
                                        'math.f\nmath.asin\n'

    # Complete math.a|angle <tab> s ...<enter> to math.asin|angle
    qtbot.keyClicks(code_editor, 'math.aangle')
    for i in range(len('angle')):
        qtbot.keyClick(code_editor, Qt.Key_Left)

    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)
        qtbot.keyPress(code_editor, 's')
    assert "asin(x)" in [x['label'] for x in sig.args[0]]
    qtbot.keyPress(completion, Qt.Key_Enter, delay=300)
    for i in range(len('angle')):
        qtbot.keyClick(code_editor, Qt.Key_Right)

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    assert code_editor.toPlainText() == 'import math\nmath.hypot\n'\
                                        'math.hypot()\nmath.asin\n'\
                                        'math.f\nmath.asin\n'\
                                        'math.asinangle\n'

    # Check math.a <tab> <backspace> <escape> do not emit sig_show_completions
    qtbot.keyClicks(code_editor, 'math.a')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    try:
        with qtbot.waitSignal(completion.sig_show_completions,
                              timeout=5000) as sig:
            qtbot.keyPress(code_editor, Qt.Key_Tab)
            qtbot.keyPress(code_editor, Qt.Key_Backspace)
            qtbot.keyPress(code_editor, Qt.Key_Escape)
        raise RuntimeError("The signal should not have been received!")
    except pytestqt.exceptions.TimeoutError:
        pass

    try:
        with qtbot.waitSignal(completion.sig_show_completions,
                              timeout=5000) as sig:
            qtbot.keyPress(code_editor, Qt.Key_Tab)
            qtbot.keyPress(code_editor, Qt.Key_Return)
        raise RuntimeError("The signal should not have been received!")
    except pytestqt.exceptions.TimeoutError:
        pass

    assert code_editor.toPlainText() == 'import math\nmath.hypot\n'\
                                        'math.hypot()\nmath.asin\n'\
                                        'math.f\nmath.asin\n'\
                                        'math.asinangle\n'\
                                        'math.\n'
    code_editor.toggle_automatic_completions(True)
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@pytest.mark.skipif(os.name == 'nt', reason='Fails on Windows')
def test_code_snippets(completions_codeeditor, qtbot):
    code_editor, completion_plugin = completions_codeeditor
    completion = code_editor.completion_widget
    snippets = code_editor.editor_extensions.get('SnippetsExtension')

    CONF.set('completions', 'enable_code_snippets', True)
    completion_plugin.after_configuration_update([])

    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(True)

    # Set cursor to start
    code_editor.go_to_line(1)

    text = """
    def test_func(xlonger, y1, some_z):
        pass
    """
    text = textwrap.dedent(text)

    code_editor.insert_text(text)
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    qtbot.keyClicks(code_editor, 'test_')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    if len(sig.args[0]) > 1:
        qtbot.keyPress(completion, Qt.Key_Tab)

    assert 'test_func(xlonger, y1, some_z)' in {
        x['label'] for x in sig.args[0]}

    expected_insert = 'test_func(${1:xlonger}, ${2:y1}, ${3:some_z})$0'
    insert = sig.args[0][0]
    assert expected_insert == insert['insertText']

    assert snippets.is_snippet_active
    assert code_editor.has_selected_text()

    # Rotate through snippet regions
    cursor = code_editor.textCursor()
    arg1 = cursor.selectedText()
    assert 'xlonger' == arg1
    assert snippets.active_snippet == 1

    qtbot.keyPress(code_editor, Qt.Key_Tab)
    cursor = code_editor.textCursor()
    arg2 = cursor.selectedText()
    assert 'y1' == arg2
    assert snippets.active_snippet == 2

    qtbot.keyPress(code_editor, Qt.Key_Tab)
    cursor = code_editor.textCursor()
    arg2 = cursor.selectedText()
    assert 'some_z' == arg2
    assert snippets.active_snippet == 3

    qtbot.keyPress(code_editor, Qt.Key_Tab)

    assert not snippets.is_snippet_active

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)

    qtbot.keyClicks(code_editor, 'test_')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    if len(sig.args[0]) > 1:
        qtbot.keyPress(completion, Qt.Key_Tab)

    # Replace selection
    qtbot.keyClicks(code_editor, 'arg1')
    qtbot.wait(5000)

    # Snippets are disabled when there are no more left
    for _ in range(0, 3):
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert not snippets.is_snippet_active

    cursor = code_editor.textCursor()
    cursor.movePosition(QTextCursor.StartOfBlock)
    cursor.movePosition(QTextCursor.EndOfBlock, mode=QTextCursor.KeepAnchor)
    text1 = cursor.selectedText()
    assert text1 == 'test_func(arg1, y1, some_z)'

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)

    qtbot.keyClicks(code_editor, 'test_')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    if len(sig.args[0]) > 1:
        qtbot.keyPress(completion, Qt.Key_Tab)

    qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert snippets.active_snippet == 2

    # Extend text from right
    qtbot.keyPress(code_editor, Qt.Key_Right, delay=300)
    qtbot.keyClicks(code_editor, '_var')

    qtbot.keyPress(code_editor, Qt.Key_Up, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Down, delay=300)

    cursor = code_editor.textCursor()
    cursor.movePosition(QTextCursor.StartOfBlock)
    cursor.movePosition(QTextCursor.EndOfBlock, mode=QTextCursor.KeepAnchor)
    text1 = cursor.selectedText()
    assert text1 == 'test_func(xlonger, y1_var, some_z)'

    cursor.movePosition(QTextCursor.EndOfBlock)
    code_editor.setTextCursor(cursor)

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)

    qtbot.keyClicks(code_editor, 'test_')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    if len(sig.args[0]) > 1:
        qtbot.keyPress(completion, Qt.Key_Tab)

    for _ in range(0, 2):
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    assert snippets.active_snippet == 3

    # Extend text from left
    qtbot.keyPress(code_editor, Qt.Key_Left, delay=300)
    qtbot.keyClicks(code_editor, 's')

    qtbot.keyPress(code_editor, Qt.Key_Tab)

    assert not snippets.is_snippet_active

    cursor = code_editor.textCursor()
    cursor.movePosition(QTextCursor.StartOfBlock)
    cursor.movePosition(QTextCursor.EndOfBlock, mode=QTextCursor.KeepAnchor)
    text1 = cursor.selectedText()
    assert text1 == 'test_func(xlonger, y1, ssome_z)'

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)

    qtbot.keyClicks(code_editor, 'test_')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    if len(sig.args[0]) > 1:
        qtbot.keyPress(completion, Qt.Key_Tab)

    assert snippets.active_snippet == 1

    # Delete snippet region
    qtbot.keyPress(code_editor, Qt.Key_Left, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Backspace, delay=300)
    assert len(snippets.snippets_map) == 3

    qtbot.keyPress(code_editor, Qt.Key_Tab)
    cursor = code_editor.textCursor()
    arg1 = cursor.selectedText()
    assert 'some_z' == arg1

    # Undo action
    with qtbot.waitSignal(code_editor.sig_undo,
                          timeout=10000) as sig:
        code_editor.undo()
    assert len(snippets.snippets_map) == 4

    for _ in range(0, 2):
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    cursor = code_editor.textCursor()
    arg1 = cursor.selectedText()
    assert 'some_z' == arg1

    with qtbot.waitSignal(code_editor.sig_redo,
                          timeout=10000) as sig:
        code_editor.redo()
    assert len(snippets.snippets_map) == 3

    for _ in range(0, 3):
        qtbot.keyPress(code_editor, Qt.Key_Tab)
    qtbot.keyPress(code_editor, Qt.Key_Right)

    qtbot.keyPress(code_editor, Qt.Key_Enter)
    qtbot.keyPress(code_editor, Qt.Key_Backspace)

    qtbot.keyClicks(code_editor, 'test_')
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    if len(sig.args[0]) > 1:
        qtbot.keyPress(completion, Qt.Key_Tab)

    # Delete text
    qtbot.keyPress(code_editor, Qt.Key_Left, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Right, delay=300)
    qtbot.keyPress(code_editor, Qt.Key_Backspace)

    for _ in range(0, 3):
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    cursor = code_editor.textCursor()
    cursor.movePosition(QTextCursor.StartOfBlock)
    cursor.movePosition(QTextCursor.EndOfBlock, mode=QTextCursor.KeepAnchor)
    text1 = cursor.selectedText()
    assert text1 == 'test_func(longer, y1, some_z)'

    CONF.set('completions', 'enable_code_snippets', False)
    completion_plugin.after_configuration_update([])

    code_editor.toggle_automatic_completions(True)
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_completion_order(completions_codeeditor, qtbot):
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget

    code_editor.toggle_automatic_completions(False)

    # Set cursor to start
    code_editor.go_to_line(1)
    qtbot.keyClicks(code_editor, 'impo')
    with qtbot.waitSignal(
            code_editor.completions_response_signal, timeout=30000):
        code_editor.document_did_change()

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(completion, Qt.Key_Tab)

    first_completion = sig.args[0][0]
    assert first_completion['insertText'] == 'import'

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(code_editor, Qt.Key_Enter, delay=300)
    qtbot.keyClicks(code_editor, 'Impo')

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(completion, Qt.Key_Tab)

    first_completion = sig.args[0][0]
    assert first_completion['insertText'] == 'ImportError'


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_text_snippet_completions(completions_codeeditor, qtbot):
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget

    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    # Set cursor to start
    code_editor.go_to_line(1)

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyClicks(code_editor, 'f')
        qtbot.keyPress(code_editor, Qt.Key_Tab, delay=300)

    results = [x for x in sig.args[0] if x['provider'] == 'Snippets']

    # Assert all retrieved words start with 'f'
    assert all({x['sortText'][1] in {'for', 'from'} for x in results})

    code_editor.toggle_automatic_completions(True)
    code_editor.toggle_code_snippets(True)


@pytest.mark.order(1)
@flaky(max_runs=5)
@pytest.mark.skipif(os.name == 'nt', reason='Hangs on Windows')
def test_completions_extra_paths(completions_codeeditor, qtbot, tmpdir):
    """Exercise code completion when adding extra paths."""
    code_editor, completion_plugin = completions_codeeditor
    completion = code_editor.completion_widget
    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    # Create a file to use as extra path
    temp_content = '''
def spam():
    pass
'''
    CONF.set('pythonpath_manager', 'spyder_pythonpath', [])
    completion_plugin.after_configuration_update([])
    qtbot.wait(500)
    qtbot.keyClicks(code_editor, 'import foo')
    qtbot.keyPress(code_editor, Qt.Key_Enter)
    qtbot.keyClicks(code_editor, 'foo.s')
    code_editor.document_did_change()
    qtbot.keyPress(code_editor, Qt.Key_Tab)
    qtbot.wait(500)
    assert code_editor.toPlainText() == 'import foo\nfoo.s'

    p = tmpdir.mkdir("extra_path")
    extra_paths = [str(p)]
    p = p.join("foo.py")
    p.write(temp_content)

    # Set extra paths
    print(extra_paths)
    CONF.set('pythonpath_manager', 'spyder_pythonpath', extra_paths)
    completion_plugin.after_configuration_update([])
    code_editor.document_did_change()
    qtbot.wait(500)

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    qtbot.keyPress(completion, Qt.Key_Tab)

    assert "spam()" in [x['label'] for x in sig.args[0]]
    assert code_editor.toPlainText() == 'import foo\nfoo.spam'

    # Reset extra paths
    CONF.set('pythonpath_manager', 'spyder_pythonpath', [])
    completion_plugin.after_configuration_update([])
    qtbot.wait(500)


@pytest.mark.order(1)
@flaky(max_runs=20)
@pytest.mark.skipif(not is_anaconda(), reason='Requires conda to work')
@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
@pytest.mark.skipif(
    not sys.platform.startswith('linux'), reason="Works reliably on Linux"
)
def test_completions_environment(completions_codeeditor, qtbot, tmpdir):
    """
    Exercise code completions when using another Jedi environment, i.e. a
    different Python interpreter.
    """
    code_editor, completion_plugin = completions_codeeditor
    completion = code_editor.completion_widget
    code_editor.toggle_automatic_completions(False)
    code_editor.toggle_code_snippets(False)

    # Get jedi test env
    py_exe = get_list_conda_envs()['Conda: jedi-test-env'][0]
    assert os.path.isfile(py_exe)

    # Check that we can't code complete Flask in the default interpreter
    # because it doesn't have it.
    qtbot.keyClicks(code_editor, 'import flas', delay=40)
    qtbot.keyPress(code_editor, Qt.Key_Tab)
    qtbot.wait(2000)
    assert code_editor.toPlainText() == 'import flas'

    # Set interpreter that has Flask and check we can provide completions for
    # it
    code_editor.set_text('')
    set_executable_helper(completion_plugin, py_exe)
    completion_plugin.after_configuration_update([])
    qtbot.wait(5000)

    qtbot.keyClicks(code_editor, 'import flas', delay=40)
    qtbot.wait(2000)
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000) as sig:
        qtbot.wait(2000)
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    assert "flask" in [x['label'] for x in sig.args[0]]
    assert code_editor.toPlainText() == 'import flask'

    set_executable_helper(completion_plugin)
    completion_plugin.after_configuration_update([])
    qtbot.wait(5000)


@pytest.mark.order(1)
@flaky(max_runs=5)
def test_dot_completions(completions_codeeditor, qtbot):
    """
    Test that completions after a dot are working as expected.

    This is a regression test for issue spyder-ide/spyder#20285
    """
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget
    code_editor.toggle_code_snippets(False)

    # Import module and check completions are shown for it after writing a dot
    # after it
    qtbot.keyClicks(code_editor, "import math")
    qtbot.keyPress(code_editor, Qt.Key_Enter)

    qtbot.wait(500)
    assert not completion.isVisible()

    with qtbot.waitSignal(completion.sig_show_completions, timeout=10000):
        qtbot.keyClicks(code_editor, "math.")

    qtbot.wait(500)
    assert completion.isVisible()

    # Select a random entry in the completion widget
    entry_index = random.randint(0, 30)
    inserted_entry = completion.completion_list[entry_index]['insertText']
    for _ in range(entry_index):
        qtbot.keyPress(completion, Qt.Key_Down, delay=50)

    # Insert completion and check that the inserted text is the expected one
    qtbot.keyPress(completion, Qt.Key_Enter)
    qtbot.wait(500)
    assert code_editor.toPlainText() == f'import math\nmath.{inserted_entry}'


@pytest.mark.order(1)
@pytest.mark.parametrize(
    "filename", ['000_test.txt', '.hidden', 'any_file.txt', 'abc.py',
                 'part.0.parquet'])
def test_file_completions(filename, mock_completions_codeeditor, qtbot):
    """
    Test that completions for files are handled as expected.

    This includes a regression test for issue spyder-ide/spyder#20156
    """
    code_editor, mock_response = mock_completions_codeeditor
    completion = code_editor.completion_widget

    # Set text to complete and move cursor to the position we want to ask for
    # completions.
    if filename == 'any_file.txt':
        # This checks if we're able to introduce file completions as expected
        # for any file when requesting them inside a string.
        qtbot.keyClicks(code_editor, "''")
    elif filename == 'abc.py':
        # This checks that we can insert file completions correctly after a
        # dot
        qtbot.keyClicks(code_editor, "'abc.'")
    elif filename == 'part.0.parquet':
        # This checks that we can insert file completions next to a dot when a
        # filename has several dots.
        qtbot.keyClicks(code_editor, "'part.0.'")
    else:
        qtbot.keyClicks(code_editor, f"'{filename[0]}'")
    code_editor.moveCursor(QTextCursor.PreviousCharacter)
    qtbot.wait(500)

    mock_response.side_effect = lambda lang, method, params: {'params': [{
        'label': f'{filename}',
        'kind': CompletionItemKind.FILE,
        'sortText': (0, f'a{filename}'),
        'insertText': f'{filename}',
        'data': {'doc_uri': path_as_uri(__file__)},
        'detail': '',
        'documentation': '',
        'filterText': f'{filename}',
        'insertTextFormat': 1,
        'provider': 'LSP',
        'resolve': True
    }]} if method == CompletionRequestTypes.DOCUMENT_COMPLETION else None

    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000):
        qtbot.keyPress(code_editor, Qt.Key_Tab, delay=300)

    qtbot.wait(500)
    assert code_editor.get_text_with_eol() == f"'{filename}'"


@pytest.mark.order(1)
@pytest.mark.parametrize(
    "directory",
    [
         pytest.param(
             '/home',
             marks=pytest.mark.skipif(
                 not sys.platform.startswith('linux'),
                 reason='Only works on Linux'
             )
        ),
        pytest.param(
            'C:\\Users',
            marks=pytest.mark.skipif(
                not os.name == 'nt',
                reason='Only works on Windows'
            )
        ),
        pytest.param(
            'C:\\Windows\\System32',
            marks=pytest.mark.skipif(
                not os.name == 'nt',
                reason='Only works on Windows'
            )
        ),
        pytest.param(
            '/Library/Frameworks',
            marks=pytest.mark.skipif(
                not sys.platform == 'darwin',
                reason='Only works on macOS'
            )
        )
    ]
)
def test_directory_completions(directory, completions_codeeditor, qtbot):
    """
    Test that directory completions work as expected.
    """
    code_editor, _ = completions_codeeditor
    completion = code_editor.completion_widget

    qtbot.wait(500)
    assert not completion.isVisible()

    if directory == '/home':
        qtbot.keyClicks(code_editor, "'/'")
    elif directory == 'C:\\Users':
        qtbot.keyClicks(code_editor, r"'C:\\'")
    elif directory == 'C:\\Windows\\System32':
        qtbot.keyClicks(code_editor, r"'C:\\Windows\\'")
    else:
        qtbot.keyClicks(code_editor, "'/Library/'")

    code_editor.moveCursor(QTextCursor.PreviousCharacter)
    with qtbot.waitSignal(completion.sig_show_completions,
                          timeout=10000):
        qtbot.keyPress(code_editor, Qt.Key_Tab, delay=300)

    qtbot.wait(500)
    assert completion.isVisible()

    # Select the corresponding entry in the completion widget
    selected_entry = False
    while not selected_entry:
        item = completion.currentItem()
        label = item.data(Qt.AccessibleTextRole).split()[0]
        if directory.split(os.sep)[-1] in label:
            selected_entry = True
        else:
            qtbot.keyPress(completion, Qt.Key_Down, delay=50)

    # Insert completion and check that the inserted text is the expected one
    qtbot.keyPress(completion, Qt.Key_Enter)
    qtbot.wait(500)
    assert osp.normpath(code_editor.toPlainText()) == f"'{directory}{os.sep}'"


if __name__ == '__main__':
    pytest.main(['test_introspection.py', '--run-slow'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Tests for gotoline.py"""

# Third party imports
from qtpy.QtWidgets import QDialogButtonBox, QLineEdit

# Local imports
from spyder.plugins.editor.widgets.gotoline import GoToLineDialog


def test_gotolinedialog_has_cancel_button(codeeditor, qtbot, tmpdir):
    """
    Test that GoToLineDialog has a Cancel button.

    Test that a GoToLineDialog has a button in a dialog button box and that
    this button cancels the dialog window.
    """
    editor = codeeditor
    dialog = GoToLineDialog(editor)
    qtbot.addWidget(dialog)
    ok_button = dialog.findChild(QDialogButtonBox).button(QDialogButtonBox.Ok)
    cancel_button = dialog.findChild(QDialogButtonBox).button(QDialogButtonBox.Cancel)
    assert not ok_button.isEnabled()
    with qtbot.waitSignal(dialog.rejected):
        cancel_button.click()


def test_gotolinedialog_enter_plus(codeeditor, qtbot):
    """
    Regression test for spyder-ide/spyder#12693
    """
    editor = codeeditor
    dialog = GoToLineDialog(editor)
    qtbot.addWidget(dialog)
    ok_button = dialog.findChild(QDialogButtonBox).button(QDialogButtonBox.Ok)
    cancel_button = dialog.findChild(QDialogButtonBox).button(QDialogButtonBox.Cancel)
    lineedit = dialog.findChild(QLineEdit)
    lineedit.setText('+')

    # Check + sign being cleared and ok button still is disabled
    lineedit.setText("+")
    assert lineedit.text() == ""
    assert not ok_button.isEnabled()

def test_gotolinedialog_check_valid(codeeditor, qtbot):
    """
    Check ok button enabled if valid text entered
    """
    editor = codeeditor
    dialog = GoToLineDialog(editor)
    qtbot.addWidget(dialog)
    ok_button = dialog.findChild(QDialogButtonBox).button(QDialogButtonBox.Ok)
    lineedit = dialog.findChild(QLineEdit)
    lineedit.setText("1")
    assert lineedit.text() == "1"
    assert ok_button.isEnabled()
    with qtbot.waitSignal(dialog.accepted):
        ok_button.click()
    assert dialog.get_line_number() == 1